#ifndef ARCHPARSER_H
#define ARCHPARSER_H 1

#include <string.h>
#include <stdio.h>
#include <math.h>
#include <algorithm>
#include "util.h"
#include "vpr_types.h"
#include <vector>
#include <string>
#include "Netlist.h"

/* This source file reads in the architectural description of an FPGA.       *
 * A # symbol anywhere in the input file denotes a comment to the end        *
 * of the line.  Put a \ at the end of a line if you want to continue        *
 * a command across multiple lines.   Non-comment lines are in the           *
 * format keyword value(s).  The entire file should be lower case.           *
 * The keywords and their arguments are:                                     *
 *                                                                           *
 *   io_rat integer (sets the number of io pads which fit into the           *
 *                  space one CLB would use).                                *
 *   chan_width_io float   (Width of the channels between the pads and       *
 *                          core relative to the widest core channel.)       *
 *   chan_width_x [gaussian|uniform|pulse] peak <width> <xpeak> <dc>.        *
 *       (<> bracketed quantities needed only for pulse and gaussian.        *
 *       Width and xpeak values from 0 to 1.  Sets the distribution of       *
 *       tracks for the x-directed channels.)                                *
 *       Other possibility:  delta peak xpeak dc.                            *
 *   chan_width_y [gaussian|uniform|pulse] peak <width> <xpeak> <dc>.        *
 *       (Sets the distribution of tracks for the y-directed channels.)      *
 *   outpin class: integer [top|bottom|left|right] [top|bottom|left|         *
 *       right] ...                                                          *
 *       (Sets the class to which each pin belongs and the side(s) of        *
 *       CLBs on which the physical output pin connection(s) is (are).       *
 *       All pins with the same class number are logically equivalent        *
 *       -- such as all the inputs of a LUT.  Class numbers must start       *
 *       at zero and be consecutive.)                                        *
 *   inpin class: integer <global> [top|bottom|left|right] [top|bottom|left| *
 *       right] ...                                                          *
 *       (All parameters have the same meanings as their counterparts        *
 *       in the outpin statement, except the optional global keyword.  If    *
 *       global is specified, this inpin can connect only to global nets.    *
 *       Global nets are not routed by the router, and are normally only     *
 *       things like clock nets.  A global pin has no switches connecting it *
 *       to the regular routing, so it takes up no area in the area model.)  *
 *                                                                           *
 *   NOTE:  The order in which your inpin and outpin statements appear       *
 *      must be the same as the order in which your netlist (.net)           *
 *      file lists the connections to the clbs.  For example, if the         *
 *      first pin on each clb in the netlist file is the clock pin,          *
 *      your first pin statement in the architecture file must be            *
 *      an inpin statement defining the clock pin.                           *
 *                                                                           *
 *   subblocks_per_clb <int>  (Number of LUT + ff logic blocks in each clb   *
 *      at most).                                                            *
 *   subblock_lut_size <int>  (Number of inputs to each LUT in the           *
 *      clbs.  Each subblock has subblock_lut_size inputs, one output        *
 *      and a clock input.)                                                  *
 *                                                                           *
 *  The following parameters only need to be in the architecture             *
 *  file if detailed routing is going to be performed (i.e. route_type       *
 *  == DETAILED).                                                            *
 *                                                                           *
 *   Fc_type [absolute|fractional]  (Are the 3 Fc values absolute            *
 *      numbers of tracks to connect to, or the fraction of the W            *
 *      tracks to which each pin can connect?)                               *
 *   Fc_output float (Sets the value of Fc -- the number of tracks           *
 *      each pin can connect to in each channel bordering the pin --         *
 *      for output pins.  The Fc_output value used is always                 *
 *      min(W,Fc_selected), so set Fc to be huge if you want Fc = W.)        *
 *   Fc_input float (Sets the value of Fc for input pins.)                   *
 *   Fc_pad float (Sets the value of Fc for pads.)                           *
 *   switch_block_type [subset|wilton|universal] (Chooses the type of        *
 *      switch block used.  See vpr_types.h for details.)                    *
 *   segment frequency: <float> length: <int | longline> wire_switch: <int>  *
 *      opin_switch: <int> Frac_cb: <float> Frac_sb: <float> Rmetal: <float> *
 *      Cmetal: <float>                                                      *
 *      Describes one type of segment.  wire_switch is the type of           *
 *      switch used when going *to* a segment of this type from any CHANX or *
 *      CHANY routing segment.  opin_switch is the type of switch used       *
 *      by a clb or pad output driver (OPIN) to connect to segments of this  *
 *      type.  Cmetal is the capacitance per logic block spanned (i.e. per   *
 *      channel segment) of a routing track of this segment type.            *
 *      Similarly, Rmetal is the resistance per logic block spanned.         *
 *   switch  <int> buffered: {yes|no} R: <float> Cin: <float>                *
 *      Cout: <float> Tdel: <float>.  Describes a type of switch.            *
 *   R_minW_nmos <float>   Resistance, in Ohms, of a minimum width nmos      *
 *      transistor.  Used only in the transistor-level area model.           *
 *   R_minW_pmos <float>   Resistance, in Ohms, of a minimum width pmos      *
 *      transistor.  Used only in the transistor-level area model.           *
 *                                                                           *
 * The following parameters allow timing analysis.                           *
 *   C_ipin_cblock <float>: Input capacitance of the buffer isolating a      *
 *      routing track from the input pin Cboxes connected to it at each      *
 *      (i,j) location.                                                      *
 *   T_ipin_cblock <float>: Delay to go through a connection block to a      *
 *      logic block (clb) input pin.                                         *
 *   T_sblk_opin_to_sblk_ipin <float>: Delay through the local interconnect  *
 *      (muxes, wires or whatever) in a clb containing multiple subblocks.   *
 *      That is, the delay from a subblock output to the input of another    *
 *      subblock in the same clb.                                            *
 *   T_clb_ipin_to_sblk_ipin <float>: Delay from a clb input pin to any      *
 *      subblock input pin (e.g. the mux delay in an Altera 8K clb).         *
 *   T_sblk_opin_to_clb_opin <float>: Delay from a subblock output to a clb  *
 *      output.  Will be 0 in many architectures.                            *
 *   T_ipad:  Delay through an input pad.                                    *
 *   T_opad:  Delay through an output pad (setup time if you assume the      *
 *               outputs are registered before being sent out).              *
 *   T_subblock T_comb: <float>  T_seq_in: <float> T_seq_out: <float>        *
 *      The combinational delay through a subblock combinational mode), the  *
 *      delay from subblock input to data being latched (including setup     *
 *      time), and the delay from clock to data coming out of the subblock   *
 *      output (i.e. clk_to_Q plus any delays due to muxes, etc.).  If a     *
 *      subblock is being used in combinational mode (clock input is open)   *
 *      then Tcomb is used to find the delay through it.  If the subblock    *
 *      is being used in sequential mode (clock input not open) then         *
 *      T_seq_in gives the delay to the storage element and T_seq_out gives  *
 *      the delay from storage element to subblock output.  You need one     *
 *      of these T_subblock lines for each of the subblocks_per_clb          *
 *      subblocks in your architecture.  The first line gives the delays for *
 *      subblock 0 (first one listed for each logic block in the netlist     *
 *      file) and so on.                                                     */


/******************* Defines and types local to this module *****************/

#define NUM_REQUIRED 8    /* Number of parameters that are always required. */
#define NUM_DETAILED 9  /* Number needed only if detailed routing used. */
#define NUM_TIMING 8   /* Number needed only if timing analysis used.  */

#define DETAILED_START NUM_REQUIRED
#define TIMING_START (NUM_REQUIRED + NUM_DETAILED)

/* Total number of different parameters in arch file. */
#define NUMINP (NUM_REQUIRED + NUM_DETAILED + NUM_TIMING)


/******************** Variables local to this module. **********************/
class Architecture {
    public:
        /** these are all you need to use for placement **/
        int nx, ny; // the width, and height in clbs of the architecture.
        int io_rat; // the number of ios which fit into the amount of space a clb takes up.
        int pins_per_clb; // the number of pins on a clb ... the maximum number of nets it could have
        std::string filename; // the arch file.

        // only public because netlist uses them.
        t_subblock_data subblock_data_ptr;
        std::vector<s_class> class_inf;
        std::vector<int> clb_pin_class;
    private:
        int num_class;
        static const char *names[NUMINP];
        int isread[NUMINP];
        t_segment_inf *segment_inf;
        t_timing_inf timing_inf_ptr;
        std::vector<int> chan_width_x;
        std::vector<int> chan_width_y;
        std::vector<bool> is_global_clb_pin;
        std::vector< std::vector<s_clb> > clb;
        std::vector<std::vector<int> > pinloc;
        std::vector<s_switch_inf> switch_inf;

    public:
        Architecture(std::string arch_file);
        ~Architecture();

        // because of the dependencies between netlist and architecture this
        // needs to be called after the netlist is created.
        void init_arch(float aspect_ratio, bool user_sized, Netlist& np);

  private:
    void countpass (FILE *fp_arch, enum e_route_type route_type,
        t_segment_inf **segment_inf_ptr, struct s_det_routing_arch
        *det_routing_arch_ptr, t_timing_inf *timing_inf_ptr);

    int get_class (FILE *fp_arch, char *buf);

    void get_pin (char *ptr, int pinnum, enum e_pin_type type,
                FILE *fp_arch, char *buf);

    enum e_Fc_type get_Fc_type (char *ptr, FILE *fp_arch, char *buf);

    enum e_switch_block_type get_switch_block_type (FILE *fp_arch,
            char *buf);

    void get_T_subblock (FILE *fp_arch, char *buf, t_T_subblock
        *T_subblock);

    void get_segment_inf (FILE *fp_arch, char *buf, t_segment_inf *seg_ptr,
         int num_switch, enum e_route_type route_type);

    void get_switch_inf (FILE *fp_arch, char *buf, int num_switch,
         enum e_route_type route_type);

    void load_global_segment_and_switch (struct s_det_routing_arch *
            det_routing_arch, t_segment_inf *segment_inf, t_timing_inf
            *timing_inf);

    void load_extra_switch_types (struct s_det_routing_arch* det_routing_arch, t_timing_inf *timing_inf);

    void check_keyword (FILE *fp, char *buf, const char *keyword);

    char *get_middle_token (FILE *fp, char *buf);

    char *get_last_token (FILE *fp, char *buf);

    int get_int (char *ptr, int inp_num, FILE *fp_arch, char *buf, int min_val);

    float get_one_float (char *ptr, int inp_num, float low_lim, float upp_lim, FILE *fp_arch, char *buf);

    float get_float (char *ptr, int inp_num, float low_lim, float upp_lim, FILE *fp_arch, char *buf);

    void get_chan (char *ptr, t_chan *chan, int inp_num, FILE *fp_arch,
        char *buf);

    void check_arch (const char *arch_file, enum e_route_type route_type,
        struct s_det_routing_arch det_routing_arch, t_segment_inf *segment_inf,
        t_timing_inf timing_inf, int max_subblocks_per_block,
        t_chan_width_dist chan_width_dist);

    void print_arch (const char *arch_file, enum e_route_type route_type,
       struct s_det_routing_arch det_routing_arch, t_segment_inf *segment_inf,
       t_timing_inf timing_inf, t_subblock_data subblock_data,
       t_chan_width_dist chan_width_dist);

    void fill_arch (void);
};

#endif
