/**
 * @copyright University of Guelph 2013
 */

#ifndef BIPARTITION_H_
#define BIPARTITION_H_ 1

#include <cstddef>

#include <iterator>
#include <list>

#include "Block.h"

namespace starplacer {
    /**
     * Two sub-partitions are the result of a split
     * of a partition, these are the values returned by
     * high_partition() and low_partition() they are not
     * created by a user.
     * This class is for
     * the recursive bi-partitioning legalization of
     * placements algorithm.
     * It defines a region or rectangular area of the
     * placement and recursively cuts the region into
     * halves which have an equal number of blocks on
     * either side.
     */
    class Bipartition {
        public:
            class iterator {
                public:
                    iterator(Bipartition *tree = NULL);
                    Bipartition& operator*();
                    iterator& operator++();
                    bool operator==(const iterator& rhs);
                    bool operator!=(const iterator& rhs);
                private:
                    Bipartition *tree_;
                    std::list<Bipartition*> stack;
            };

            Bipartition(int top, int bottom, int left, int right, int io_ratio);

            /**
             * Set the points for the region to partition,
             * these points are passed down to children so
             * this method is only for the root.
             */
            template <class InnerIterator, class OutterIterator>
                void LoadPoints(
                        InnerIterator first_inner,
                        InnerIterator last_inner,
                        OutterIterator first_outter,
                        OutterIterator last_outter) {
                assert(root() == this);
                assert(inner_x_.empty());
                assert(inner_y_.empty());
                assert(outter_x_.empty());
                assert(outter_y_.empty());

                while (first_inner != last_inner) {
                    PointData* p = new PointData(*first_inner);
                    inner_x_.push_back(p);
                    inner_y_.push_back(p);
                    //assert(contains(p->x(), p->y()));
                    ++first_inner;
                }

                while (first_outter != last_outter) {
                    PointData* p = new PointData(*first_outter);
                    outter_x_.push_back(p);
                    outter_y_.push_back(p);
                    //assert(contains(p->x(), p->y()));
                    ++first_outter;
                }

                assert(inner_x_.size() == inner_y_.size());
                assert(outter_x_.size() == outter_y_.size());
                Sort();
            }

            ~Bipartition();


            void Partition();

            /**
             * defines the directions for which a partition can be split.
             * kSplitNone == 0 to allow for nicer if statements
             * `if (partition.split_direction())`
             */
            enum SplitDirection {kSplitNone = 0,
                                kSplitHorizontal,
                                kSplitVertical};

            int area() const;
            int CLBCols();
            int CLBRows();
            int top() const;
            int bottom() const;
            int left() const;
            int right() const;
            int width() const;
            int height() const;
            int inner_count() const;
            int outter_count() const;
            /**
             * The width of the region
             */
            //int width() const;

            void Sort();
            /**
             * The height of the region
             */
           // int height() const;

            void dumpPartitions(std::string tag);

            bool contains(int x, int y);

            int inner_point_count() const;

            int outter_point_count() const;

            int id() const;

            void dump(std::string tag);

            /**
             * Remove all points and partitions
             */
            void clear();

            /**
             * The direction in which this region was split (if at all).
             */
            SplitDirection split_direction() const;

            
            /**
            * the io ratio from the architecture
            */
            int outter_ratio() const;

            /**
             * Move all Points in this region to be at the centre of the region.
             */
            void PutAllToCentre();

            /**
             * Scale and translate the points in the region to fill the region.
             * As a result of calling this, the minimum x value will be the x origin,
             * the minimum y value will be the y origin  and the maximum values
             * will be at the edges.
             */
            void FitToRegion();

            /**
             * Recursively split the region
             * for a given number of iterations or until there is at most 1 block in each region.
             * use the favour_split ratio to break ties.
             * @param iterations do at most this many iterations, negative values indicate no limit.
             */
            void Subdivide(SplitDirection tie_breaker, int iterations=-1);

            /**
             * Computes the mean average and variance values
             * for the x and y positions of the blocks in the
             * region. you may pass NULL to any which you are not
             * interested in.
             */
            void InnerVariance(double* avg_x, double* avg_y,
                    double* variance_x, double* variance_y) const;

            /**
             * Retrieves the partition with the values
             * lower (or equal to) the split_value
             * or NULL if the partition has not been split.
             */
            Bipartition* low_partition() const;

            /**
             * Retrieves the partition with the values
             * higher (or equal to) the split_value
             * or NULL if the partition has not been split.
             */
            Bipartition* high_partition() const;

            /**
             * Get the root partition which defines the boundaries
             * between inner and outter positions.
             */
            Bipartition* root();

            /**
             * Asks the root to split this region.
             */
            void Split(SplitDirection direction);

            /**
             *
             * Splits the region into two sub regions divided by the mid point
             * in the y-axis
             */
            void SplitVertical();

            /**
             * Splits the region into two sub regions divided by the mid point
             * in the x-axis
             */
            void SplitHorizontal();

            iterator begin();

            iterator end();

            /**
             * This class wraps the given positions in another class
             * to allow for easy attachment of data to the positions
             * which can be used by the algorithms.
             */
            class PointData {
                public:
                    explicit PointData(Block* p);
                    bool is_low;  // set to avoid sorting both the regions.
                    double x() const;
                    double y() const;
                    void set_x(double x);
                    void set_y(double y);
                    void move_to(double x, double y);
                private:
                    Block* point;
            };

        protected:

            typedef std::vector<PointData*> PointContainer;

            int top_, bottom_, left_, right_;
            double pos_x_, pos_y_;
            int max_inner_, max_outter_;
            int min_inner_, min_outter_;
            int split_value_;
            int outter_ratio_;
            int id_;
            SplitDirection split_dir_;
            PointContainer inner_x_;
            PointContainer inner_y_;
            PointContainer outter_x_;
            PointContainer outter_y_;
            Bipartition* low_sub_;
            Bipartition* high_sub_;
            Bipartition* root_;

        private:
            void init(int x, int y, int width, int height, int io_ratio, Bipartition *root, int id);
            Bipartition(int x, int y, int width, int height, int io_ratio, Bipartition *root, int id);
            void InitSubpartitions();
            void split(
                Bipartition::PointContainer& from_x,
                Bipartition::PointContainer& from_y,
                Bipartition::PointContainer& low_x,
                Bipartition::PointContainer& high_x,
                Bipartition::PointContainer& low_y,
                Bipartition::PointContainer& high_y,
                int min_low, int max_low, double split_x,
                bool compare_x, double* split_centre);
    };
}  // namespace starplacer

#endif  // BIBPARTITION_H_
