/**
 * @copyright University of Guelph
 */

#ifndef __CONJUGATE_GRADIENT_H__
#define __CONJUGATE_GRADIENT_H__ 1

#include <string>
#include <vector>
#include <list>

#include "Placement.h"
#include "Block.h"
#include "IOBlock.h"
#include "ConjugateGradient.h"

namespace starplacer {
    class ConjugateGradient {
      private:
        class CGData {
            public:
            double dx, dy; // derivative
            double rx, ry; // residuals
            double mrx, mry; // modified residuals

            // IOBlocks only.. (could use a CGData subclass)
            double dxy;
            double dxx, dyy;
            
            /**
            * set everything to 0.
            */
            void Initialize();
        };

        std::vector<CGData> block_data_;
        double delta;
        double err;
        Placement* place;

      public:
        typedef std::list<Block*> VarList;
        typedef std::list<Block*>::iterator variable_iterator;
        typedef CGData CGBlockData;
        typedef CGData CGIOData;
        
        std::list<Block*> variables; // the variables to act on.
        // some blocks could be constant for a time, and then added
        // back as variables.
        
        double beta; // how much the direction vector changes from
        // old direction toward the new direction.
        
        double alpha; // how far in the direction vector a block is moved.

        ConjugateGradient(Placement* p);
        
        /**
        * set all fields of internal CGData structs to 0.
        */
        void InitializeData();
        
        /**
        * After adding variables this method acts on the placement
        * and changes the block positions.
        */
        void Go(int count, double epsilon);
        
        /**
        * Add the blocks as variables in the linear system
        * At first the io are constants and so they are not
        * included in the first few iterations. @see CGPlacer.
        */
        template<class iterator>
        void AddVariables(iterator start, iterator end) {
            variables.insert(variables.end(), start, end);
        }
        
        /**
        * Dump the CGData fields to a file for debugging.
        */
        void dump_data(std::string filename);
        
    private:
        void Iterate();
        
        /**
        * a Helper function to calculate the residuals for
        * a single block.
        */
        void CalculateResiduals(Block* block);
        
        /**
        * IOblocks have some sort of optimization applied
        * to their residuals.
        */
        void ModifyResiduals(IOBlock* block, double *mx, double *my);

        /**
        * updates the residuals (and modified residuals)
        * for each block, and calculates the internal beta
        * value, the beta value controls how much the direction
        * vector changes between iterations.
        */ 
        void CalculateResidualsAndBeta();
        
        /**
        * Update the internal Alpha value.
        * the alpha value controls how far
        * a block is moved in the direction of its gradient.
        */
        double CalculateAlpha();
        
        /**
        * get the Gradient data associated with the block
        */
        inline CGBlockData* block_data(const Block& block) {
            assert(block.id() < (int)block_data_.size());
            return &block_data_[block.id()];
        }

        /**
        * Get the Gradient data associated with the I/O
        * block.
        */
        inline CGIOData* io_data(const IOBlock& block) {
            assert(block.id() < (int)block_data_.size());
            return &block_data_[block.id()];
        }
   };
} // namespace starplacer

#endif
