/**
 * Copyright 2013, University of Guelph
 */

#ifndef INCLUDE_XTREE_H_
#define INCLUDE_XTREE_H_ 1


#include <assert.h>
#include <cstddef>

#include <list>

namespace starplacer {

    /**
     * @class DoubleEndedPriorityQueue
     * @brief A binary tree in which you can eXtract the eXtreme values.
     * @author starplacer Ming
     * The DoubleEndedPriorityQueue class is used to as part of the Successive-Over-Relaxtion analytic placement method.
     * The tree is used to efficently store ans sort the variables so to help with choosing
     * the optimal order to evaluate them in.
     */
    template <typename T>
    class DoubleEndedPriorityQueue {
        public:
            /**
             * A tree is made up of nodes which are a key value pair.
             */
            class Node {
                friend class DoubleEndedPriorityQueue;
                public:

                    /**
                     * Creates a Node, which wraps the object to be placed in the tree.
                     */
                    explicit Node(T obj_id);

                    /**
                     * Returns the object attached to this node.
                     */
                    T id() const;

                    /**
                     * Returns the key value that is used in sorting the tree.
                     */
                    double key() const;

                    /**
                     * Used in debugging to recursively count the number of nodes
                     * in a subtree.
                     */
                    int Count() const;

                    /**
                     * Locates the node in the tree tied to the object
                     * @note assumes that the object is unique in the tree
                     */
                    typename DoubleEndedPriorityQueue<T>::Node* Find(T obj_id);

                protected:
                    T id_;
                    double key_;  // the key value to sort by
                    Node* parent_;  // if NULL assumes not in the tree or root of the tree.
                    Node* left_;  // a node with a lesser key value or NULL
                    Node* right_;  // a node with a greater or equal key value or NULL
            };

            /**
             * Iterates through the nodes in the tree in a
             * pre-order traversal. Only used for debugging.
             */
            class iterator {
                public:
                    iterator(Node *t = NULL);
                    Node& operator*();
                    iterator& operator++();
                    bool operator==(const iterator& rhs);
                    bool operator!=(const iterator& rhs);
                private:
                    Node *tree_;
                    std::list<Node*> stack_;
            };

      public:
        /**
        * Creates a tree.
        */
        DoubleEndedPriorityQueue<T>();

        /**
        * Returns the number of nodes currently in the tree
        */
        int size() const;

        /**
        * @brief Inserts a node into the tree, the node must not already be in a tree.
        * @param n The node to store (cannot be NULL)
        * @param key The value to order the node by.
        */
        void Insert(Node* n, double key);

        /**
        * @brief Removes and returns the node with the minimum key value in the tree.
        * @return The node with the minimum key or NULL is the tree is empty.
        */
        Node* ExtractMin();

        /**
        * @brief Removes and returns the node with the maximum key value in the tree.
        * @return The node with the maximum key or NULL is the tree is empty.
        */
        Node* ExtractMax();

        /**
        * Remove the node from the Tree.
        * @note does not free the Node
        */
        Node* Remove(Node* del);

        /**
        * Change the key value for the node, moving it in the tree and updating
        * its key variable.
        */
        void set_key(Node* obj_id, double key);

        /**
        * @brief Search recursively for a node with the object attached.
        * @return The node with obj_id attached or NULL if none found.
        */
        Node* Find(T obj_id) const;

        /**
        * Checks if the nodes is in the tree.
        * @param n The Node to check for.
        * @return True if the node is in the tree, false otherwise.
        * @warn assumes the node cannot belong to any other tree.
        */
        bool Contains(Node* n) const;

      private:
        Node* root_;  // NULL if empty
        int size_; // number of nodes in the tree.
    };

    /* Node definition */
    /*
     * It looks like an AVL tree.
     * only one instance allowed (see Contains)
     */
    template <typename T>
    DoubleEndedPriorityQueue<T>::Node::Node(T obj_id) {
        id_ = obj_id;
        left_ = right_ = parent_ = NULL;
    }

    template <typename T>
    double DoubleEndedPriorityQueue<T>::Node::key() const {
        return key_;
    }

    template <typename T>
    T DoubleEndedPriorityQueue<T>::Node::id() const {
        return id_;
    }

    template <typename T>
    typename DoubleEndedPriorityQueue<T>::Node* DoubleEndedPriorityQueue<T>::Find(T obj_id) const {
        if (root_) {
            return root_->Find(obj_id);
        } else {
            return NULL;
        }
    }

    template <typename T>
    typename DoubleEndedPriorityQueue<T>::Node* DoubleEndedPriorityQueue<T>::Node::Find(T obj_id) {
        if (id_ == obj_id) {
            return this;
        }

        if (left_) {
            typename DoubleEndedPriorityQueue<T>::Node* n = left_->Find(obj_id);
            if (n) return n;
        }

        if (right_) {
            return right_->Find(obj_id);
        }
        return NULL;
    }

    template <typename T>
    int DoubleEndedPriorityQueue<T>::Node::Count() const {
        int count = 1;

        if (left_) {
            count += left_->Count();
        }

        if (right_) {
            count += right_->Count();
        }

        return count;
    }

    /* iterator definition */
    template <typename T>
    DoubleEndedPriorityQueue<T>::iterator::iterator(DoubleEndedPriorityQueue<T>::Node* t) {
        tree_ = t;
    }

    template <typename T>
    typename DoubleEndedPriorityQueue<T>::Node& DoubleEndedPriorityQueue<T>::iterator::operator*() {
        return *tree_;
    }

    template <typename T>
    typename DoubleEndedPriorityQueue<T>::iterator& DoubleEndedPriorityQueue<T>::iterator::operator++() {
        if (tree_->left_) stack_.push_back(tree_->left_);
        if (tree_->right_) stack_.push_back(tree_->right_);
        if (stack_.empty()) {
            tree_ = NULL;
        } else {
            tree_ = stack_.back();
            stack_.pop_back();
        }
        return *this;
    }

    template <typename T>
    bool DoubleEndedPriorityQueue<T>::iterator::operator!=(const DoubleEndedPriorityQueue<T>::iterator& rhs) {
        return !(*this == rhs);
    }

    template <typename T>
    bool DoubleEndedPriorityQueue<T>::iterator::operator==(const DoubleEndedPriorityQueue<T>::iterator& rhs) {
        return (rhs.tree_ == tree_ && rhs.stack_ == stack_);
    }

    /* DoubleEndedPriorityQueue definition */
    template <typename T>
    DoubleEndedPriorityQueue<T>::DoubleEndedPriorityQueue() {
        root_ = NULL;
        size_ = 0;
    }

    template <typename T>
    int DoubleEndedPriorityQueue<T>::size() const {
        return size_;
    }

    template <typename T>
    void DoubleEndedPriorityQueue<T>::Insert(Node* n, double key) {
        assert(n);
        assert(n->parent_ == NULL);

        n->key_ = key;
        n->left_ = n->right_ = NULL;
        size_++;

        if (root_ == NULL) {
            root_ = n;
            n->parent_ = NULL;
            return;
        }

        int level = 0;
        Node* p = root_;

        while (p != NULL) {
            n->parent_ = p;
            if (key > p->key_) {
                p = p->right_;
            } else {
                p = p->left_;
            }

            level++;
        }

        p = n->parent_;
        if (key > p->key_) {
            p->right_ = n;
        } else {
            p->left_ = n;
        }

        if (1 << (level >> 1) < size())  // checking if balanced
            return;

        Node* grandp = NULL;

        while (p != NULL) {
            grandp = p->parent_;
            if (n == p->right_) {
                p->right_ = n->left_;
                if (p->right_ != NULL) {
                    p->right_->parent_ = p;
                }
                n->left_ = p;
            } else {
                p->left_ = n->right_;
                if (p->left_ != NULL) {
                    p->left_->parent_ = p;
                }
                n->right_ = p;
            }

            p->parent_ = n;
            n->parent_ = grandp;

            if (grandp != NULL) {
                if (p == grandp->right_) {
                    grandp->right_ = n;
                } else {
                    grandp->left_ = n;
                }
            }
            p = grandp;
        }

        root_ = n;
    }

    template <typename T>
    typename DoubleEndedPriorityQueue<T>::Node* DoubleEndedPriorityQueue<T>::ExtractMin() {
        if (root_ == NULL) {
            return NULL;
        }

        Node* min = root_;
        while (min->left_ != NULL) {
            min = min->left_;
        }

        Remove(min);
        return min;
    }

    template <typename T>
    typename DoubleEndedPriorityQueue<T>::Node* DoubleEndedPriorityQueue<T>::ExtractMax() {
        if (root_ == NULL) {
            return NULL;
        }

        Node* max = root_;
        while (max->right_ != NULL) {
            max = max->right_;
        }

        Remove(max);
        return max;
    }

    template <typename T>
    typename DoubleEndedPriorityQueue<T>::Node* DoubleEndedPriorityQueue<T>::Remove(Node* del) {
        Node* p = del->parent_;
        assert(p || del == root_);

        Node* rep = del->left_;
        if (rep != NULL) {
            while (rep->right_ != NULL) {
                rep = rep->right_;
            }

            if (rep->parent_ == del) {
                del->left_ = rep->left_;
            } else {
                rep->parent_->right_ = rep->left_;
            }

            if (rep->left_ != NULL) {
                rep->left_->parent_ = rep->parent_;
            }
        } else {
            rep = del->right_;
            if (rep != NULL) {
                while (rep->left_ != NULL) {
                    rep = rep->left_;
                }

                if (rep->parent_ == del) {
                    del->right_ = rep->right_;
                } else {
                    rep->parent_->left_ = rep->right_;
                }

                if (rep->right_ != NULL) {
                    rep->right_->parent_ = rep->parent_;
                }
            }
        }

        if (rep != NULL) {
            rep->left_ = del->left_;
            rep->right_ = del->right_;
            rep->parent_ = p;

            if (rep->left_ != NULL) {
                rep->left_->parent_ = rep;
            }

            if (rep->right_ != NULL) {
                rep->right_->parent_ = rep;
            }

            if (p != NULL) {
                if (del == p->left_) {
                    p->left_ = rep;
                } else {
                    p->right_ = rep;
                }
            }

            if (root_ == del) {
                root_ = rep;
            }
        } else if (p == NULL) {
            root_ = NULL;
        } else if (del == p->left_) {
            p->left_ = NULL;
        } else {
            p->right_ = NULL;
        }

        del->parent_ = del->left_ = del->right_ = NULL;
        size_--;
        return del;
    }

    template <typename T>
    void DoubleEndedPriorityQueue<T>::set_key(Node* obj_id, double key) {
        if (key == obj_id->key()) {
            return;
        }

        Remove(obj_id);
        Insert(obj_id, key);
    }

    template <typename T>
    bool DoubleEndedPriorityQueue<T>::Contains(Node* n) const {
        assert(n);
        return n->parent_ != NULL || n == root_;
    }
}  // namespace starplacer

#endif  // INCLUDE_XTREE_H_
