/**
 * @copyright University of Guelph
 */
#ifndef _NET_H_
#define _NET_H_ 1

#include <vector>
#include <string>

#include "yaml.h"

namespace starplacer {
    class Placement;

    class Net {
        friend class Placement;
        int id_; // a unique contiguous id
        std::vector<Block*> blocks_; // the blocks connected to this net.

  public:
    static const double alpha_factor; //! a multiplier used to make star+ values similar to HPWL
    double x, y; //! the centre of the net, or the average of the block positions connected to it.
    double sumx, sumy, sumx2, sumy2; //! $f( \sigma_{b \in Net}{b_x}, \sigma_{b \in Net}{b_y} $f)
    double costx, costy; // star+ cost in the x or y dimension
    bool is_global;

    /**
    * num_pins is the cardinality of the net.
    * The id is the unique identifier, the
    * index in the netlist of VPR's structure.
    * @see Netlist.net
    */
    Net(int num_pins, int id);

    /**
    * does nothing useful
    */
    void CondenseEdges();

    inline Block* block(int i) {
        return blocks_[i];
    }

    inline size_t num_blocks() const {
        return blocks_.size();
    }

    /**
    * A unique id from [0 .. netlist.num_nets)
    * is assigned to each net, use this for creating
    * lookups for data to associate with this net.
    */
    inline int id() const {
        return id_;
    }

    /* dump the attributes to the console */
    void print() const;

    /**
    * Adds a block to the net, should be used in conjunction
    * with Block::addNet(), the pin_index is the pin
    * of the block which the net connects to.
    */
    void addBlock(Block* block);

    /**
     * Computes the Star+ cost in the X dimension
     * from scratch. 
     * @warn Netlist::alpha_factor is not applied
     */
    void CalculateXCost();

    /**
     * Computes the Star+ cost in the Y dimension
     * from scratch.
     * @warn Netlist::alpha_factor is not applied
     */
    void CalculateYCost();

    /**
    * Used to update the Star+ cost without
    * recomputing from scratch.
    */
    void BlockMoved(double oldx, double oldy, double newx, double newy);
};
} // namespace starplacer


namespace yaml {
    std::string yaml(starplacer::Net* net);
}

#endif
