#ifndef NETPARSER_H
#define NETPARSER_H 1

#include <string.h>
#include <stdio.h>

#include "vpr_types.h"
#include "util.h"
#include "hash.h"

#include <assert.h>
#include <vector>
#include <string>

/* This source file reads in a .net file.  A .net file is a netlist   *
 * format defined by me to allow arbitary logic blocks (clbs) to      *
 * make up a netlist.  There are three legal block types: .input,     *
 * .output, and .clb.  To define a block, you start a line with one   *
 * of these keywords; the function of each type of block is obvious.  *
 * After the block type keyword, you specify the name of this block.  *
 * The line below must start with the keyword pinlist:, and gives a   *
 * list of the nets connected to the pins of this block.  .input and  *
 * .output blocks are pads and have only one net connected to them.   *
 * The number of pins on a .clb block is specified in the arch-       *
 * itecture file, as are the equivalences between pins.  Each         *
 * clb must have the same number of pins (heterogeneous FPGAs are     *
 * currently not supported) so any pins which are not used on a clb   *
 * must be identified with the reserved word "open".  All keywords    *
 * must be lower case.                                                *
 *                                                                    *
 * The lines immediately below the pinlist line must specify the      *
 * contents of the clb.  Each .subblock line lists the name of the    *
 * subblock, followed by the clb pin number or subblock output to     *
 * which each subblock pin should connect.  Each subblock is assummed *
 * to consist of a LUT with subblock_lut_size inputs, a FF, and an    *
 * output.  The pin order is input1, input2, ... output, clock.  The  *
 * architecture file sets the number of subblocks per clb and the LUT *
 * size used.  Subblocks are used only for timing analysis.  An       *
 * example clb declaration is:                                        *
 *                                                                    *
 * .clb name_of_clb  # comment                                        *
 *  pinlist:  net_1 net_2 my_net net_of_mine open D open              *
 *  subblock: sub_1 0 1 2 3 open open open                            *
 *  subblock: sub_2 1 3 ble_0 open open 5 open                        *
 *                                                                    *
 * Notice that the output of the first subblock (ble_0) is used only  *
 * by the second subblock.  This is fine.                             *
 *                                                                    *
 * Ending a line with a backslash (\) means it is continued on the    *
 * line below.  A sharp sign (#) indicates the rest of a line is      *
 * a comment.                                                         *
 * The vpack program can be used to convert a flat blif netlist       *
 * into .net format.                                                  *
 *                                                                    *
 * V. Betz, Jan. 29, 1997.                                            */

/* A note about the way the character buffer, buf, is passed around. *
 * strtok does not make a local copy of the character string         *
 * initially passed to it, so you must make sure it always exists.   *
 * Hence, I just use one buffer declared in read_net and pass it     *
 * downstream.  Starting a new line with an automatic variable       *
 * buffer in a downstream subroutine and then returning and trying   *
 * to keep tokenizing it would cause problems since the buffer now   *
 * lies on a stale part of the stack and can be overwritten.         */

class Architecture;

class Netlist {
  public:
    std::vector<struct s_net> net;
    std::vector<struct s_block> block;
    std::vector<bool> is_global;

    int num_nets;
    int num_blocks;
    int num_globals;
    int num_p_inputs;
    int num_p_outputs;
    int num_clbs;
    std::string filename;
  private:
    /* Temporary storage used during parsing. */
    int *num_driver, *temp_num_pins;
    struct s_hash **hash_table;
    int temp_block_storage;

    /* Used for memory chunking of everything except subblock data. */

    int chunk_bytes_avail;
    char *chunk_next_avail_mem;

    /* Subblock data can be accessed anywhere within this module.  Pointers to *
     * the main subblock data structures are passed back to the rest of the    *
     * program through the subblock_data_ptr structure passed to read_net.     */

    int max_subblocks_per_block;
    int subblock_lut_size;
    t_subblock **subblock_inf;
    int *num_subblocks_per_block;

    /* The subblock data is put in its own "chunk" so it can be freed without  *
     * hosing the other netlist data.                                          */

    int ch_subblock_bytes_avail;
    char *ch_subblock_next_avail_mem;
    struct s_linked_vptr *ch_subblock_head_ptr;
    Architecture& arch;

  public:
    Netlist(std::string net_file, Architecture& arch);

    ~Netlist();

    void init_parse(int doall);

    char *get_tok (char *buf, int doall, FILE *fp_net);

    int get_pin_number (char *ptr);

    void load_subblock_array (int doall, FILE *fp_net, char *temp_buf, int num_subblocks, int bnum);

    void set_subblock_count (int bnum, int num_subblocks);

    char *parse_subblocks (int doall, FILE *fp_net, char *buf,
               int bnum);

    char *add_clb (int doall, FILE *fp_net, char *buf);

    void add_io (int doall, int block_type, FILE *fp_net, char *buf);

    void parse_name_and_pinlist (int doall, FILE *fp_net, char *buf);

    void add_global (int doall, FILE *fp_net, char *buf);

    int add_net (char *ptr, enum e_pin_type type, int bnum, int blk_pnum,
          int doall);

    /* Release memory needed only during circuit netlist parsing. */
    void free_parse (void);
};

#endif
