/**
 *  @copyright University of Guelph
 *  @file Shrubbery.h
 *  @brief Shrubbery is a pre-placement algorithm which places I/O blocks.
 *
 * Shrubbery works by using Djikstra's shortest path algorithm to find the blocks which are
 * most connected to the I/O blocks and eventually which I/O blocks should be placed near
 * to one another. They are then placed along the outside of the board.
 */

#ifndef __SHRUBBERY_H__
#define __SHRUBBERY_H__ 1

#include <assert.h>

#include <list>
#include <vector>

#include "Placement.h"
#include "Block.h"
#include "IOBlock.h"
#include "Net.h"

#include "SetPartition.h"
#include "DoubleEndedPriorityQueue.h"

namespace starplacer {

    class Shrubbery {
      private:
        typedef std::list<IOBlock*> Cycle;
        typedef Cycle::iterator cycle_iterator;
        typedef std::vector<IOBlock*>::iterator shrub_iterator;
        typedef DoubleEndedPriorityQueue<Net*> IOHeap;
        typedef std::vector<IOHeap::Node*> HeapNodes;

        std::vector<IOBlock*> shrubs_; // Maps a block id to is shrub represented by the IOBlock at the root of the shrub.
        std::vector<IOBlock*> net_shrubs_; // Maps a net to a shrub root
        std::vector<Cycle> cycle_memory_; // A single lump of memory allocated once to make freeing it easier.
        std::vector<Cycle*> cycles_; // each shrub has a cycle which grows as it is combined with other shrubs
        // the cycle is the order in which the combined shrubs (IOblocks) should be placed relative to one another.
        std::vector<SetPartition> partitions_; // Stores whether or not two ioblocks have combined to create a larger shrub.
        size_t shrub_count; // the number of remaining shrubs, as they expand and combine this will reduce to 1.
        HeapNodes tnodes_; // These nodes each wrap an IOBlock
        IOHeap tree; // The heap using in Djikstra's shortest path to find the next most connected net/IO to a shrub
        Placement* place; // the placement to pre-place on.
        std::vector<IOBlock*> io_sequence; // represents the final sequence of ios, after all shrubs are combined and ordered.

      public:
        ~Shrubbery();

        /**
         * This constructor runs shrubbery, the caller needs nothing more
         * than to pass in the placement and the end result is that IO blocks
         * are placed around the border and all other blocks are placed at
         * the centre of the FPGA.
         */
        Shrubbery(Placement* placement);

        /**
         * A partition is used to associate shrubs as a bigger shrub when
         * they are combined. Basically, all shrubs that have been combined
         * are from the same partition.
         */
        inline SetPartition* partition(const IOBlock& b) {
            assert(b.io_id < (int)partitions_.size());
            return &partitions_[b.io_id];
        }

        /**
         * Return the shrub (root) associated with
         * this net or NULL if it doesn't belong to any shrub.
         */
        inline IOBlock*& shrub(const Net& n) {
            assert(n.id() < (int)net_shrubs_.size());
            return net_shrubs_[n.id()];
        }

        /**
         * Lookup the node in the heap to update its closeness to
         * a block in one of the shrubs.
         */
        inline Shrubbery::IOHeap::Node*& tnode(const Net& n) {
            assert(n.id() < (int)tnodes_.size());
            return tnodes_[n.id()];
        }

        /**
         * Lookup the shrub root for the block (NULL)
         * if no shrub is associated with the block yet.
         */
        inline IOBlock*& shrub(const Block& b) {
            assert(b.id() < (int)shrubs_.size());
            return shrubs_[b.id()];
        }

        /**
         * Lookup the cycle associated with the shrub root.
         */
        inline Shrubbery::Cycle*& cycle(const IOBlock& b) {
            assert(b.id() < (int)cycles_.size());
            return cycles_[b.id()];
        }

        /**
         * For each IO representing a shrub, expand the shrub to include the most connected net to any block
         * currently in the shrub.  When two shrubs meet place their i/o blocks next to each other by combining
         * their cycles and then combine the shrubs into one. Continue until only 1 giant shrub remains, with a
         * cycle containing all the ios.
         */
        void ExpandShrubs();

        /**
         * Given two small sequences of IO blocks this method merges the blocks in an attempt to reduce
         * the distance between two highly connected blocks.
         */
        void MergeCycles(IOBlock& block, IOBlock& other);

        /**
         * Extract the io_sequence from the shrub data
         */
        void SequenceIO();

        /**
         * Place the I/Os in the I/O sequence around the border of the placement, spaces them as evenly as possible.
         */
        void WrapIOSequence();

        /**
         * Place all the blocks (including IO) in the placement at the centre of the FPGA
         */
        void CentreBlocks();

   }; // class Shrubbery
} // namespace starplacer
 #endif
