/**
 * @copyright University of Guelph
 */

#ifndef __SIMPLE_BIPARTITION_H__
#define __SIMPLE_BIPARTITION_H__ 1

#include <vector>

#include "Block.h"
#include "IOBlock.h"
#include "DoubleEndedPriorityQueue.h"
#include "Placement.h"

namespace starplacer {

    // Defines a rectangular area of the FPGA, according to the global architecture
    // where top is the lower y coordinate, and l is the lower x coordinate
    // the maximum number of clbs and ios (according to the area and the architecture are set here)
    // the minimum number is set relative to the another region in initSub
    // if the region is a column or row vector of ios only, then the region will shrink to exclude the corner pieces.
    // this makes sure that regions are always rectangular by allowing them to include a corner piece and removing
    // them when the result would still be a rectangle.

    class SimpleBipartition {
      public:
        /**
         * density is a multiplier applied to the area
         * to determine how many clb blocks are allowed in a region
         * given its area. density > 1 allows for some overlap
         * density < 1 may create impossible to satisfy conditions.
         * density = 1 fully legalize the placement, no overlap.
         */
        SimpleBipartition(Placement* p, double density);

        ~SimpleBipartition();

        /**
        * Partition the blocks given by initialize.
        * you must re-initialize before calling this again.
        */
        void Partition();

        /**
        * Run through the io blocks and place them in the z dimension as well.
        */
        void SpreadIO();

        /**
        * Load the clb and io blocks, these will be the ones which are partitioned.
        * It is often better to avoid partitioning isolated blocks since these
        * can be placed at the end instead of the pushing around other blocks
        * so not all blocks should always be placed.
        */
        void Initialize(std::vector<Block*> blocks, std::vector<IOBlock*> ios);

      private:
        typedef DoubleEndedPriorityQueue<Block*>::Node HeapNode;
        typedef std::vector<HeapNode*> HeapNodes; 

        SimpleBipartition* root_; // the root, or NULL if this
        // region is the root.
        Placement* placement; // placement to partition on 
        // note only blocks added to Initialize will be partitioned.

        HeapNodes tNodeX; // memory for all regions allocated by the root.
        HeapNodes tNodeY; // used in sorting

        std::vector<Block*> clbs; // clbs to be partitioned
        std::vector<IOBlock*> ios; // ios to be partitioned

        /**
        * private constructor for creating the child regions.
        */
        SimpleBipartition(SimpleBipartition* root, int t, int b, int l, int r, double density);
        
        /**
        * Get the root partition
        */
        SimpleBipartition* root();
        
        /**
        * a helper method to reduce code duplication in
        * the two constructors.
        */
        void InitRegion(int t, int b, int l, int r, double density);
        
        /**
        * 
        */
        void partitionIO();
        
        /**
        *
        */
        void partitionY();
        
        /**
        *
        */
        void partitionX();
        
        /**
        * After a split the subregions have some dependant variables.
        * Mostly the restrictions of the minimum and maximum number
        * of clbs / io blocks the region can hold, the maximum is
        * defined as the size of the area - the minimum the other
        * region can hold.
        */
        void initSub();
        
        /**
        * Place all the blocks in the region at the centre.
        * Near the end of the partitioning all the regions
        * are a single CLB region so it is essentially
        * placing them in the clb / io  position.
        * this does not place io blocks in the z dimension.
        * Placing IO in the z dimension is done by SpreadIO.
        */
        void position();
        
        /**
        * a helper method.
        * Move blocks from a split it the x direction
        * from one region to its sister region.
        */
        void moveSubCLBx(SimpleBipartition* from, SimpleBipartition* to);
        
        /**
        * a helper method.
        * Move blocks from a split it the y direction
        * from one region to its sister region.
        */
        void moveSubCLBy(SimpleBipartition* from, SimpleBipartition* to);

        SimpleBipartition* sub1; // child subregions.
        SimpleBipartition* sub2;

        int minclbs, maxclbs, minios, maxios; // split constraints
        double density_; // maimum allowed density >= 1.
        double x, y; // x a y split points in terms ob block positions (not region space)
        int top, bottom, left, right; // the boundaries of the region
    };
}

#endif
