/**
 * @copyright University of Guelph
 */

#ifndef __SUCCESSIVE_OVER_RELAXATION_H__
#define __SUCCESSIVE_OVER_RELAXATION_H__ 1

#include <list>
#include <vector>
#include <string>

#include "Block.h"
#include "IOBlock.h"
#include "Net.h"
#include "Placement.h"
#include "DoubleEndedPriorityQueue.h"
#include "yaml.h"

namespace starplacer {

    class SuccessiveOverRelaxation {

      public:
        class SORNode;
        class BlockNode;
        class IONode;
        class NetNode;

        typedef std::vector<NetNode*> NetNodes;
        typedef std::vector<BlockNode*> BlockNodes;
        typedef DoubleEndedPriorityQueue<SORNode*>::Node HeapNode;

        typedef std::list<SORNode*> VarSequence;
        typedef std::vector<BlockNode*> VariableSequence;

        void AddToSequence(Block* block); // add a block to the sequence
        // call this after CombineAndOrderSequences
        
        // update all the nets costs.
        // this is slightly different than the Net::CalculateCostX
        // and Y updates.
        void UpdateNets();

        /**
        * The below classes wrap blocks and nets
        * into a common hierarchy so that they
        * can all be treated as nodes in a graph.
        */
        class SORNode {
          public:
            double weight; // weight used in shortest path for 'closeness'
            double omega; // the relaxation factor
            int LPindex; // the index in the variable ordering (-1 is the invalid index)
            HeapNode* heap_node; // used in Djikstra's shortest path to
            // find the next closest node when calculating the order of nodes
            // to evaluate in.
            
            /**
            * set the position of the underlying block/net
            */
            virtual void setPosition(double x, double y) = 0;
            
            /**
            * how many nodes (net/block) uis this connected to.
            */
            virtual int num_nodes() = 0;
            
            /**
            * Get the i'th node connected to this node,
            * the block_nodes and net_nodes are passed in so that it
            * can pick out the corresponding node.
            */
            virtual SORNode* node(int i, BlockNodes& block_nodes, NetNodes& net_nodes) = 0;
            
            /**
            * Update the position of the node.
            * i.e. move the x and y position of the block/net it represents
            * @param ome the omega to use for this calculation
            * @param quick use the LPindex to avoid calculating out of order.
            * @param placement the placement
            * @param netnodes see below
            * @param blocknodes see below
            * the class needs to look up other information for the nodes it is connected to.
            * it seems best if it is oblivois to the SuccessiveOverRelaxation class as much as
            * possible so these are passed in as arguments,
            */
            virtual void CalculatePosition(double ome, bool quick, Placement* placement, BlockNodes& block_nodes, NetNodes& net_nodes) = 0;
            
            /**
            * calculate the weight of the node
            * to be used in Djikstra's Shortest path algorithm
            * for finding closely connected blocks.
            * updates the weight instance variable.
            */
            virtual void CalculateWeight() = 0;
            
            /**
            * Recalculate omega (the relaxation factor) for the block.
            */
            virtual void CalculateOmega(BlockNodes& block_nodes) = 0;
            
            /**
            * returns the weight in the closeness graph of the
            * edge between this node and another.
            */
            virtual double EdgeWeight(SORNode* node) = 0;
            
            /**
            * print the instance variables
            * to the console for debugging
            */
            virtual void dump() = 0;
        };

        class BlockNode: public SORNode {
          public:
            Block* block;
            BlockNode();
            virtual ~BlockNode();
            virtual void dump();
            virtual void setPosition(double x, double y);
            virtual int num_nodes();
            virtual SORNode* node(int i, BlockNodes& block_nodes, NetNodes& net_nodes);
            virtual void CalculateOmega(BlockNodes& block_nodes);
            void CalculateTimingPosition(double* x_pos, double* y_pos);
            virtual void CalculatePosition(double ome, bool quick, Placement* placement, BlockNodes& block_nodes, NetNodes& net_nodes);
            virtual void CalculatePosition(Placement* placement);
            virtual void CalculateWeight();
            virtual double EdgeWeight(SORNode* node);
        };

        class NetNode: public SORNode {
          public:
            Net* net;
            NetNode();
            virtual ~NetNode();
            virtual void dump();
            virtual void setPosition(double x, double y);
            virtual int num_nodes();
            virtual SORNode* node(int i, BlockNodes& block_nodes, NetNodes& net_nodes);
            virtual void CalculateOmega(BlockNodes& block_nodes);
            virtual void CalculatePosition(double ome, bool quick, Placement* placement, BlockNodes& block_nodes, NetNodes& net_nodes);
            virtual void CalculateWeight();
            virtual double EdgeWeight(SORNode* node);
        };

        class IONode: public BlockNode {
          public:
            virtual void dump();
            static double IOWeight;
            virtual void CalculateWeight();
            virtual void CalculatePosition(Placement* placement);
            virtual void CalculatePosition(double ome, bool quick, Placement* placement, BlockNodes& block_nodes, NetNodes& net_nodes);
        };

        std::vector<BlockNode*> blocks_;
        std::vector<NetNode*> nets_;

        typedef std::vector<NetNode*>::iterator netnode_iterator;
        typedef std::vector<BlockNode*>::iterator blocknode_iterator;

      public:
        VarSequence sequence;
        Placement* placement;

        double cost();
        SuccessiveOverRelaxation(Placement* placement);
        ~SuccessiveOverRelaxation();

        /**
        * Returns a sequence-- a list of all the blocks
        * and net nodes in order of "closeness" to the src.
        * this is used in onrder variables for efficient evalutaion.
        */
        SuccessiveOverRelaxation::VarSequence CalculateSequence(Block* src);

        /**
        * each sequence given by CalculateSequence is a list of all the blocks
        * and net nodes in order of "closeness" to the src.
        * This method combines a bunch of sequenes and their source nodes
        * into a single sequence (where no net or list appears more than once)
        * the result is stored in the internal `sequence` variable and is
        * the one used to order the calculation of the variables.
        */
        void CombineAndOrderSequences(VarSequence& src, std::vector<VarSequence>& seqs);
        
        /**
        * Remove all the nets from the internal sequence.
        */
        void RemoveNets();

        /**
         * converge the internal sequence, and make use of the LPindex speedup.
        */   
        void QuickConverge();

        /**
        * dump the state to a file for debugging.
        */
        void dump_data(std::string filename);

        /**
        * converge all the blocks (call RemoveNets) before this
        * to reduce the variables.
        */
        void ConvergeBlock(int rounds);

        /**
        * return the node wrapping the block argument.
        */
        BlockNode* BlockLookup(Block* b);

        /**
        * return the node wrapping the net argument.
        */
        NetNode* NetLookup(Net* n);
     };
} // namespace starplacer

namespace yaml {
    std::string yaml(starplacer::SuccessiveOverRelaxation::SORNode* node);
}
#endif
