/**
 * @copyright University of Guelph
 */

#include <stdio.h>
#include <math.h>
#include <assert.h>

#include <iostream>
#include <cstddef>
#include <algorithm>
#include <limits>
#include <list>
#include <string>
#include <vector>

#include "Block.h"
#include "Bipartition.h"

namespace starplacer {
    Bipartition::iterator::iterator(Bipartition* tree) {
        tree_ = tree;
    }

    Bipartition& Bipartition::iterator::operator*() {
        return *tree_;
    }

    Bipartition::iterator& Bipartition::iterator::operator++() {
        if (tree_->split_direction() != kSplitNone) {
            stack.push_back(tree_->low_partition());
            stack.push_back(tree_->high_partition());
        }
        if (stack.empty()) {
            tree_ = NULL;
        } else {
            tree_ = stack.back();
            stack.pop_back();
        }
        return *this;
    }

    bool Bipartition::iterator::operator==(const iterator& rhs) {
        return (rhs.tree_ == tree_ && stack == rhs.stack );
    }

    bool Bipartition::iterator::operator!=(const iterator& rhs) {
        return !(*this == rhs);
    }

    Bipartition::iterator Bipartition::begin() {
        return iterator(this);
    }

    Bipartition::iterator Bipartition::end() {
        return iterator(NULL);
    }

    Bipartition::PointData::PointData(Block* p) {
        is_low = false;
        point = p;
    }

    void Bipartition::clear() {
        assert(root() == this);
        if (low_sub_) {
            delete low_sub_;
        }
        if (high_sub_) {
            delete high_sub_;
        }
        low_sub_ = NULL;
        high_sub_ = NULL;
        inner_x_.clear();
        inner_y_.clear();
        outter_x_.clear();
        outter_y_.clear();
        split_dir_ = kSplitNone;
        split_value_ = 0;
    }

    double Bipartition::PointData::x() const {
        return point->x;
    }

    void Bipartition::PointData::set_x(double x) {
        point->x = x;;
    }

    double Bipartition::PointData::y() const {
        return point->y;
    }

    void Bipartition::PointData::move_to(double x, double y) {
        point->x = x;
        point->y = y;
    }

    void Bipartition::PointData::set_y(double y) {
        point->y = y;
    }

    int Bipartition::outter_point_count() const {
        if (split_direction() == kSplitNone) {
            return outter_x_.size();
        } else {
            return low_partition()->outter_point_count() + high_partition()->outter_point_count();
        }
    }

    int Bipartition::inner_point_count() const {
        if (split_direction() == kSplitNone) {
            return inner_x_.size();
        } else {
            return low_partition()->inner_point_count() + high_partition()->inner_point_count();
        }
    }

    bool Bipartition::contains(int xx, int yy) {
        if (xx < left() || xx > right() || yy < top() || yy > bottom()) {
            return false;
        }
        return true;
    }

    Bipartition::Bipartition(int top, int bottom, int left, int right, int io_ratio) {
        init(top, bottom, left, right, io_ratio, NULL, 1);
    }

    Bipartition::Bipartition(int top, int bottom, int left, int right,
            int io_ratio, Bipartition* root, int id) {
        init(top, bottom, left, right, io_ratio, root, id);
    }

    void Bipartition::init(int top, int bottom, int left, int right, int io_ratio, Bipartition* arch, int identifier) {
        assert(top <= bottom); // weird but matches VPR
        assert(left <= right);
        assert(left >= 0);
        assert(top >= 0);
        assert(io_ratio >= 0);
        assert(arch == NULL || arch->root() == arch);

        top_ = top;
        bottom_ = bottom;
        left_ = left;
        right_ = right;
        root_ = arch;
        outter_ratio_ = io_ratio;
        id_ = identifier;

        low_sub_ = NULL;
        high_sub_ = NULL;
        split_dir_ = kSplitNone;
        min_inner_ = 0; // updated later in initsubpartitions
        min_outter_ = 0;

        const int width = CLBCols();
        const int height = CLBRows();
        max_inner_ = width * height;

        if (width <= 0) { // x is completely in IO land
            if (top_ <= root()->top())
                top_ = root()->top() + 1; // remove "top" corner from your 'area'
            if (bottom_ >= root()->bottom())
                bottom_ = root()->bottom() - 1; // remove corner
        }

        if (height <= 0) { // y is completely in IO
            if (left_ <= root()->left())
                left_ = root()->left() + 1; // remove corner
            if (right_ >= root()->right())
                right_ = root()->right() - 1;
        }

        pos_x_ = (left_ + right_) * 0.5f; // where to place the blocks
        pos_y_ = (top_ + bottom_) * 0.5f;

        max_outter_ = 0;
        if (top_ <= root()->top()) {
            max_outter_ += width;
        }
        if (bottom_ >= root()->bottom()) {
            max_outter_ += width;
        }
        if (left_ <= root()->left()) {
            max_outter_ += height;
        }
        if (right_ >= root()->right()) {
            max_outter_ += height;
        }
        max_outter_ *= io_ratio;
    }

    Bipartition::~Bipartition() {
        if (low_sub_) {
            delete low_sub_;
        }
        if (high_sub_) {
            delete high_sub_;
        }
        for (PointContainer::iterator pi = inner_x_.begin();
                pi != inner_x_.end(); ++pi)
            delete *pi;
        for (PointContainer::iterator po = outter_x_.begin();
                po != outter_x_.end(); ++po)
            delete *po;
    }

    Bipartition::SplitDirection Bipartition::split_direction() const {
        return split_dir_;
    }

    int Bipartition::right() const {
        return right_;
    }
    int Bipartition::left() const {
        return left_;
    }
    int Bipartition::top() const {
        return top_;
    }
    int Bipartition::bottom() const {
        return bottom_;
    }

    int Bipartition::CLBCols() {
        int w = width();
        if (left() <= root()->left()) --w;
        if (right() >= root()->right()) --w;
        return w;
    }

    int Bipartition::CLBRows() {
        int h = height();
        if (top() <= root()->top()) --h;
        if (bottom() >= root()->bottom()) --h;
        return h;
    }

    int Bipartition::area() const {
        return width() * height();
    }

    int Bipartition::height() const {
        return bottom_ - top_ + 1;
    }

    int Bipartition::width() const {
        return right_ - left_ + 1;
    }

    int Bipartition::outter_ratio() const {
        return outter_ratio_;
    }

    void Bipartition::dump(std::string tag) {
        return;
        printf("--- # %s\n", tag.c_str());
        int i = 0;
        for (iterator it = begin(); it != end(); ++it) {
            printf("<ix>\n");
            for (PointContainer::iterator t = (*it).inner_x_.begin();
                    t != (*it).inner_x_.end(); ++t) {
                printf("node:%d, x:%.2f, y:%.2f\n",
                        i++, (*t)->x(), (*t)->y());
            }
            printf("</ix>\n");
            printf("<iy>\n");
            for (PointContainer::iterator t = (*it).inner_y_.begin();
                    t != (*it).inner_y_.end(); ++t) {
                printf("node:%d, x:%.2f, y:%.2f\n",
                        i++, (*t)->x(), (*t)->y());
            }
            printf("</iy>\n");
            printf("<ox>\n");
           for (PointContainer::iterator t = (*it).outter_x_.begin();
                    t != (*it).outter_x_.end(); ++t) {
                printf("node:%d, x:%.2f, y:%.2f\n",
                        i++, (*t)->x(), (*t)->y());
            }
            printf("</ox>\n");
            printf("<oy>\n");
            for (PointContainer::iterator t = (*it).outter_y_.begin();
                    t != (*it).outter_y_.end(); ++t) {
                printf("node:%d x:%.2f y:%.2f\n",
                        i++, (*t)->x(), (*t)->y());
            }
            printf("</oy>\n");
        }
        printf("... # %s\n", tag.c_str());
    }

    void Bipartition::Partition() {
        double avg_x, avg_y, variance_x, variance_y;
        InnerVariance(&avg_x, &avg_y, &variance_x, &variance_y);
        Subdivide(variance_x > variance_y ? Bipartition::kSplitVertical : Bipartition::kSplitHorizontal);
        PutAllToCentre();
    }

    void Bipartition::dumpPartitions(std::string tag) {
        return;
        printf("<%s>\n", tag.c_str());
        for (iterator it = begin(); it != end(); ++it) {
            printf("top:%d bottom:%d left:%d right:%d\n",
                (*it).top(), (*it).bottom(), (*it).left(), (*it).right());
        }
        printf("</%s>\n", tag.c_str());
    }

    void Bipartition::PutAllToCentre() {
        // the choice between inner_x or inner_y is arbitrary
        for (PointContainer::iterator iter_inner = inner_x_.begin();
            iter_inner != inner_x_.end(); ++iter_inner) {
            (*iter_inner)->move_to(pos_x_, pos_y_);
        }

        for (PointContainer::iterator iter_outter = outter_x_.begin();
            iter_outter != outter_x_.end(); ++iter_outter) {
            (*iter_outter)->move_to(pos_x_, pos_y_);
        }

        if (low_sub_) {
            low_sub_->PutAllToCentre();
        }

        if (high_sub_) {
            high_sub_->PutAllToCentre();
        }
    }

    static bool compx(Bipartition::PointData* a, Bipartition::PointData* b) {
        return a->x() < b->x();
    }

    static bool compy(Bipartition::PointData* a, Bipartition::PointData* b) {
        return a->y() < b->y();
    }

    void Bipartition::Sort() {
        // Stable Sort, for comparing against C#
        std::stable_sort(inner_x_.begin(), inner_x_.end(), compx);
        std::stable_sort(inner_y_.begin(), inner_y_.end(), compy);
        std::stable_sort(outter_x_.begin(), outter_x_.end(), compx);
        std::stable_sort(outter_y_.begin(), outter_y_.end(), compy);
        // Unstable Quick-sort (faster)
        //std::sort(inner_x_.begin(), inner_x_.end(), compx);
        //std::sort(inner_y_.begin(), inner_y_.end(), compy);
        //std::sort(outter_x_.begin(), outter_x_.end(), compx);
        //std::sort(outter_y_.begin(), outter_y_.end(), compy);
    }

    int Bipartition::id() const {
        return id_;
    }

    void Bipartition::InnerVariance(double *avgx, double *avgy,
                double* varx, double* vary) const {
        double sumx = 0;
        double sumy = 0;
        double sumx2 = 0;
        double sumy2 = 0;
        double avg_x = 0;
        double avg_y = 0;

        for (PointContainer::const_iterator it = inner_x_.begin();
                it != inner_x_.end(); ++it) {
            sumx += (*it)->x();
            sumx2 += (*it)->x() * (*it)->x();
            sumy += (*it)->y();
            sumy2 += (*it)->y() * (*it)->y();
        }

        int count = inner_x_.size();
        if (count == 0) {
           if (avgx) *avgx = 0;
           if (avgy) *avgy = 0;
           if (varx) *varx = 0;
           if (vary) *vary = 0;
           return;
        }

        avg_x = sumx / count;
        avg_y = sumy / count;
        if (varx) *varx = sumx2 / count - avg_x * avg_x;
        if (vary) *vary = sumy2 / count - avg_y * avg_y;
        if (avgx) *avgx = avg_x;
        if (avgy) *avgy = avg_y;
    }

    Bipartition* Bipartition::low_partition() const {
         return low_sub_;
    }

    Bipartition* Bipartition::high_partition() const {
        return high_sub_;
    }

    void Bipartition::FitToRegion() {
        double minx, miny, maxx, maxy;

        if (!inner_x_.empty()) {
            minx = inner_x_.front()->x();
            miny = inner_y_.front()->y();
            maxx = inner_x_.back()->x();
            maxy = inner_y_.back()->y();
        } else {
            minx = std::numeric_limits<double>::max();
            miny = std::numeric_limits<double>::max();
            maxx = std::numeric_limits<double>::min();
            maxy = std::numeric_limits<double>::min();
        }

        if (!outter_x_.empty()) {
            minx = std::min(minx, outter_x_.front()->x());
            miny = std::min(miny, outter_y_.front()->y());
            maxx = std::max(maxx, outter_x_.back()->x());
            maxy = std::max(maxy, outter_y_.back()->y());
        }

        double x_span = maxx - minx;
        double y_span = maxy - miny;

        if (x_span == 0.0 || y_span == 0.0) {
            return;
        }

        const double scale_x = (right() - left()) / x_span;
        const double scale_y = (bottom() - top()) / y_span;

        for (PointContainer::iterator b = inner_x_.begin();
                b != inner_x_.end(); ++b) {
            const double tx = left() + scale_x * ((*b)->x() - minx);
            const double ty = top() + scale_y * ((*b)->y() - miny);
            (*b)->move_to(tx, ty);
            assert(contains(tx, ty));
        }

        for (PointContainer::iterator b = outter_x_.begin();
                b != outter_x_.end(); ++b) {
            const double tx = left() + scale_x * ((*b)->x() - minx);
            const double ty = top() + scale_y * ((*b)->y() - miny);
            (*b)->move_to(tx, ty);
            assert(contains(tx, ty));
        }
    }

    Bipartition* Bipartition::root() {
        return root_ ? root_ : this;
    }

    int Bipartition::inner_count() const {
        return (int)inner_x_.size();
    }

    int Bipartition::outter_count() const {
        return (int)outter_x_.size();
    }

    void Bipartition::Subdivide(SplitDirection tie_breaker, int iterations) {
        // Base case
        if (width() <= 1 && height() <= 1) {
            if (inner_count() > 0 || outter_count() > 0) {
                PutAllToCentre();
            }
            return;
        }

        if (iterations == 0) {
            return;
        } else if (iterations > 0) {
            iterations -= 1;
        }

        // split the larger dimension and alternate the split for squares
        if (width() > height()) {
            Split(kSplitVertical);
            tie_breaker = kSplitHorizontal;
        } else if (width() < height()) {
            Split(kSplitHorizontal);
            tie_breaker = kSplitVertical;
        } else if (tie_breaker == kSplitVertical) {
            Split(kSplitVertical);
            tie_breaker = kSplitHorizontal;
        } else {
            Split(kSplitHorizontal);
            tie_breaker = kSplitVertical;
        }

        // recurse
        low_sub_->Subdivide(tie_breaker, iterations);
        high_sub_->Subdivide(tie_breaker, iterations);
    }

   void Bipartition::split(
            Bipartition::PointContainer& from_main,
            Bipartition::PointContainer& from_other,
            Bipartition::PointContainer& low_main,
            Bipartition::PointContainer& high_main,
            Bipartition::PointContainer& low_other,
            Bipartition::PointContainer& high_other,
            int min_low, int max_low, double split_value,
            bool compare_x, double* split_centre) {

        assert(low_main.empty());
        assert(high_main.empty());
        assert(high_other.empty());
        assert(low_other.empty());
        assert(from_main.size() == from_other.size());

        //int comp_val = 0;
        int i = min_low;

        while (i < max_low) {
            if (compare_x && from_main[i]->x() >= split_value) {
                break;
            } else if (!compare_x && from_main[i]->y() >= split_value) {
                break;
            } else {
                ++i;
            }
        }

        for (int j = 0; j < i; ++j) {
            PointData* p = from_main[j];
            low_main.push_back(p);
            p->is_low = true;
        }

        for (int j = i; j < (int)from_main.size(); ++j) {
            PointData* p = from_main[j];
            high_main.push_back(p);
            p->is_low = false;
        }

        if (i < (int)from_main.size()) {
            if (compare_x) {
                *split_centre = from_main[i]->x();
            } else {
                *split_centre = from_main[i]->y();
            }
        }

        Bipartition::PointContainer::iterator it = from_main.begin();
        for (it = from_other.begin(); it != from_other.end(); ++it) {
            if ((*it)->is_low)
                low_other.push_back(*it);
            else
                high_other.push_back(*it);
        }

        assert(high_other.size() == high_main.size());
        assert(low_other.size() == low_main.size());
        assert(high_other.size() + low_other.size() == from_other.size());
        assert(high_main.size() + low_main.size() == from_main.size());
        from_main.clear();
        from_other.clear();
    }

    void Bipartition::InitSubpartitions() {
        if (low_sub_->contains(high_sub_->left(), high_sub_->top())) {
            dumpPartitions("PartitionBounds");
        }
        assert(!low_sub_->contains(high_sub_->left(), high_sub_->top()));
        assert(!low_sub_->contains(high_sub_->left(), high_sub_->bottom()));
        assert(!low_sub_->contains(high_sub_->right(), high_sub_->top()));
        assert(!low_sub_->contains(high_sub_->right(), high_sub_->bottom()));

        low_sub_->max_inner_ = std::min((int)inner_x_.size(), low_sub_->max_inner_);
        low_sub_->max_outter_ = std::min((int)outter_x_.size(), low_sub_->max_outter_);
        high_sub_->max_inner_ = std::min((int)inner_x_.size(), high_sub_->max_inner_);
        high_sub_->max_outter_ = std::min((int)outter_x_.size(), high_sub_->max_outter_);

        low_sub_->min_inner_ = inner_x_.size() - high_sub_->max_inner_;
        low_sub_->min_outter_ = outter_x_.size() - high_sub_->max_outter_;
        high_sub_->min_inner_ = inner_x_.size() - low_sub_->max_inner_;
        high_sub_->min_outter_ = outter_x_.size() - low_sub_->max_outter_;

        assert(low_sub_->max_inner_ >= low_sub_->min_inner_);
        assert(low_sub_->max_outter_ >= low_sub_->min_outter_);
        assert(high_sub_->max_inner_ >= high_sub_->min_inner_);
        assert(high_sub_->max_outter_ >= high_sub_->min_outter_);
    }

    void Bipartition::SplitVertical() {
        if (left() == right()) return;

        split_dir_ = kSplitVertical;
        const int middle = (left() + right()) / 2;

        low_sub_ = new Bipartition(top(), bottom(), left(), middle, outter_ratio(), root(), 2 * id());
        high_sub_ = new Bipartition(top(), bottom(), middle + 1, right(), outter_ratio(), root(), 2 * id() + 1);

        InitSubpartitions();
        double space = pos_x_;
        /** split the inner values **/
        split(inner_x_, inner_y_, low_sub_->inner_x_, high_sub_->inner_x_,
                low_sub_->inner_y_, high_sub_->inner_y_,
                low_sub_->min_inner_, low_sub_->max_inner_,
                pos_x_, true, &space);
        pos_x_ = space;
        /** split the outter values **/
        split(outter_x_, outter_y_, low_sub_->outter_x_, high_sub_->outter_x_,
                low_sub_->outter_y_, high_sub_->outter_y_,
                low_sub_->min_outter_, low_sub_->max_outter_,
                pos_x_, true, &space);
    }

    void Bipartition::SplitHorizontal() {
        /* printf("split:y w:%d h:%d clbs:%d ios:%d minio:%d maxio:%d minclb:%d maxclb:%d\n",
                width(), height(), (int)inner_x_.size(), (int)outter_x_.size(),
                min_outter_, max_outter_, min_inner_, max_inner_); */
        if (height() <= 1)
            return;

        split_dir_ = kSplitHorizontal;
        const int middle = (top() + bottom()) / 2;

        low_sub_ = new Bipartition(top(), middle, left(), right(), outter_ratio(), root(), 2 * id());
        high_sub_ = new Bipartition(middle + 1, bottom(), left(), right(), outter_ratio(), root(), 2 * id() + 1);

        InitSubpartitions();
        /** split the inner values **/
        split(inner_y_, inner_x_, low_sub_->inner_y_, high_sub_->inner_y_,
                low_sub_->inner_x_, high_sub_->inner_x_,
                low_sub_->min_inner_, low_sub_->max_inner_,
                pos_y_, false, &pos_y_);

        /** split the outter values **/
        double space;
        split(outter_y_, outter_x_, low_sub_->outter_y_, high_sub_->outter_y_,
                low_sub_->outter_x_, high_sub_->outter_x_,
                low_sub_->min_outter_, low_sub_->max_outter_,
                pos_y_, false, &space);
    }

    void Bipartition::Split(SplitDirection direction) {
        switch (direction) {
            case kSplitNone:
                return;
            case kSplitVertical:
                SplitVertical();
                break;
            case kSplitHorizontal:
                SplitHorizontal();
                break;
            default:
                throw 1;
        }
    }

}  // namespace starplacer
