/**
 * @copyright University of Guelph
 */

#include <iostream>
#include <assert.h>
#include <cstddef>
#include <math.h>

#include <algorithm>
#include <limits>
#include <list>
#include <vector>

#include "ConjugateGradient.h"
#include "Shrubbery.h"
#include "IOBlock.h"
#include "Block.h"
#include "Bipartition.h"
#include "SimpleBipartition.h"
#include "Placement.h"

#include "CGPlacer.h"

namespace starplacer {

     void CGPlacer::Place(Placement* placement) {
        std::vector<Block*> _clbs;
        std::vector<IOBlock*> _ios;

        ConjugateGradient cg(placement);
        Shrubbery shrub(placement);

        for (Placement::block_iterator b = placement->blocks.begin();
                b != placement->blocks.end(); ++b) {
            if (std::find(placement->isolated_blocks.begin(),
                        placement->isolated_blocks.end(), *b) != placement->isolated_blocks.end()) {
                continue;
            }

            IOBlock *io = dynamic_cast<IOBlock*>(*b);
            if (io) {
                _ios.push_back(io);
            } else {
                _clbs.push_back(*b);
            }
        }

        cg.AddVariables(_clbs.begin(), _clbs.end());
        cg.InitializeData();

        //cg.dump_data("cg_after_insert.csv");
        cg.Go(100, 0.1);
        //printf("Alpha: %f\nBeta:%f\n", cg.alpha, cg.beta);

        cg.AddVariables(_ios.begin(), _ios.end());
        cg.Go(1001, 0.01);

        bool use_bipart = true; // easy to try out different methods.
        
        /**
        * This is the original author's legalization method for SOR
        */
        SimpleBipartition legalize_A(placement, 1); // needed for IO placement at end.
        
        /**
        * This is the bipartitioning algorithm the original author uses.
        */
        Bipartition legalize_B(0, placement->arch.ny + 1, 0, placement->arch.nx + 1, placement->arch.io_rat);
        
        if (use_bipart) {
            legalize_B.LoadPoints(_clbs.begin(), _clbs.end(), _ios.begin(), _ios.end());
            legalize_B.Partition();
        } else {
            legalize_A.Initialize(_clbs, _ios);
            legalize_A.Partition();
        }

        int count = std::max(placement->arch.nx, placement->arch.ny);

        printf("---#  Star+ Wire-length Cost\n");
        int iter_count = 0;
        while (count > 1) {
            cg.InitializeData();
            printf(" - %d\n", (int)placement->StarPlusWirelengthCost());
            cg.Go(count, 0.01);
            count = (int)(count * 0.99);
            if (use_bipart) {
                legalize_B.clear();
                legalize_B.LoadPoints(_clbs.begin(), _clbs.end(), _ios.begin(), _ios.end());
                legalize_B.Partition();
            } else {
                legalize_A.Initialize(_clbs, _ios);
                legalize_A.Partition();
            }
            iter_count++;
        }

        legalize_A.Initialize(_clbs, _ios);
        legalize_A.SpreadIO();
        printf("%d iterations\n", iter_count);
    }
} // namespace starplacer
