/**
 * @copyright University of Guelph
 */

#include <assert.h>
#include <cstddef>
#include <math.h>
#include <stdio.h>

#include <string>
#include <vector>
#include <list>
#include <fstream>
#include <iostream>

#include "Block.h"
#include "IOBlock.h"
#include "Placement.h"

#include "ConjugateGradient.h"

namespace starplacer {
    void ConjugateGradient::CGData::Initialize() {
        // specific to IO blocks
        dxy = 0;
        dxx = 0;
        dyy = 0;

        // for all blocks
        dx = 0;
        dy = 0;
        rx = 0;
        ry = 0;
        mrx = 0;
        mry = 0;
    }

    void ConjugateGradient::dump_data(std::string filename) {
        FILE* fp = fopen(filename.c_str(), "w");
        fprintf(fp, "id,\t\tx,\t\ty,\t\tdx,\t\tdy,\t\trx,\t\try,\t\t"
                "mrx,\t\tmry,\t\tdxy,\t\tdxx,\t\tdyy\n");
        for (ConjugateGradient::VarList::iterator b = variables.begin();
                b != variables.end(); ++b) {
            Block* block = *b;
            CGBlockData* bdata = block_data(*block);
            fprintf(fp, "%d,\t%f,\t%f,\t%f,"
                    "\t%f,\t%f,\t%f,"
                    "\t%f,\t%f,\t%f,"
                    "\t%f,\t%f\n", block->id(),
                    block->x, block->y, bdata->dx,
                    bdata->dy, bdata->rx, bdata->ry,
                    bdata->mrx, bdata->mry,
                    bdata->dxy, bdata->dxx,
                    bdata->dyy);
        }
        fclose(fp);
    }

    void ConjugateGradient::CalculateResiduals(Block* block) {
        // Equation 4.12
        // rx, ry pertain to the gradient calculation
        // Equation 4.15
        // mx, my pertain to the hessian matrix in eq 4.15
        // they are missing \alpha and Sum{d_j} the gradient.
        double mx = 0, my = 0, temp;
        CGBlockData* data = block_data(*block);

        data->rx = 0;
        data->ry = 0;

        for (size_t i = 0; i < block->num_nets(); ++i) {
            Net *vn = block->nets(i);
            if (vn->is_global) {
                continue;
            }

            temp = (vn->x - block->x) / vn->costx; // d S_l / d_j
            data->rx += temp;
            mx += (1 - 1.0 / vn->num_blocks() - temp * temp) / vn->costx;

            temp = (vn->y - block->y) / vn->costy;
            data->ry += temp;
            my += (1 - 1.0 / vn->num_blocks() - temp * temp) / vn->costy;

            assert(!isnan(data->ry));
            assert(!isnan(data->rx));
            assert(!isnan(my));
            assert(!isnan(mx));
        }

        // rx is now Equation 4.12 (without alpha)
        // mx is Equation 4.11 (without alpha)
        IOBlock* io = dynamic_cast<IOBlock*>(block);
        if (io) {
            ModifyResiduals(io, &mx, &my);
        }

        data->mrx = mx == 0 ? 0 : data->rx / mx;
        data->mry = my == 0 ? 0 : data->ry / my;
        assert(!isnan(data->mry));
        assert(!isnan(data->mrx));
    }

    void ConjugateGradient::ModifyResiduals(IOBlock* block, double *mx, double *my) {
        // read the variables as: xn = x ^ n where x is normalized to the chip size, ie x1 is in [-1, 1]
        const double radx = (place->arch.nx + 1) * .5; // the ~~ radius ~~ of the chip
        const double rady = (place->arch.ny + 1) * .5; // the ~~ radius ~~ of the chip
        const double x1 = block->x / radx - 1;
        const double x2 = x1 * x1;
        const double x3 = x2 * x1;
        const double x4 = x2 * x2;
        const double y1 = block->y / rady - 1;
        const double y2 = y1 * y1;
        const double y3 = y2 * y1;
        const double y4 = y2 * y2;
        const double w = 0.001;

        CGIOData* data = io_data(*block);
        data->rx -= w * (x4 + y4 - 1) * x3 / radx;
        *mx += data->dxx = w * (7 * x4 + 3 * y4 - 3) * x2 / radx / radx;
        data->ry -= w * (x4 + y4 - 1) * y3 / rady;
        *my += data->dyy = w * (7 * y4 + 3 * x4 - 3) * y2 / rady / rady;
        data->dxy = w * 4 * x3 * y3 / radx / rady;
    }

    ConjugateGradient::ConjugateGradient(Placement* placement) {
        assert(placement);
        place = placement;
        block_data_.resize(place->blocks.size());
        delta = 0;
        beta = 0;
        err = 0;
    }

    void ConjugateGradient::Iterate() {
        double alpha = CalculateAlpha();
        double terr; // max squared (x or y) error

        typedef std::list<Block*>::iterator Iblock;

        err = 0;
        Block* block = NULL;

        for (Iblock b = variables.begin(); b != variables.end(); ++b) {
            block = *b;
            CGBlockData* bdata = block_data(*block);
            // Equation 4.1
            // update the positions
            block->x += terr = alpha * bdata->dx;
            assert(!isnan(block->x));

            terr *= terr;
            if (err < terr) {
                err = terr;
            }

            block->y += terr = alpha * bdata->dy;
            assert(!isnan(block->y));

            terr *= terr;
            if (err < terr) {
                err = terr;
            }
        }

        CalculateResidualsAndBeta();

        for (Iblock b = variables.begin(); b != variables.end(); ++b) {
            block = *b;
            assert(block);
            CGBlockData* bdata = block_data(*block);
            // Equation 4.6
            bdata->dx = bdata->mrx + beta * bdata->dx;
            bdata->dy = bdata->mry + beta * bdata->dy;
        }
    }

    void ConjugateGradient::Go(int count, double epsilon) {
        delta = 0;

        for (VarList::iterator b = variables.begin(); b != variables.end(); ++b) {
            Block* block = *b;
            CGBlockData* bdata = block_data(*block);
            CalculateResiduals(block);

            bdata->dx = bdata->mrx; // Equation 4.2
            delta += bdata->rx * bdata->mrx;
            assert(!isnan(delta));

            bdata->dy = bdata->mry;
            delta += bdata->ry * bdata->mry;
            assert(!isnan(delta));
        }

        for (int i = 0; i < count; ++i) {
            Iterate();
            if (err < epsilon) {
                break;
            }
        };
    }

    void ConjugateGradient::CalculateResidualsAndBeta() {
        for (Placement::net_iterator n = place->nets.begin(); n != place->nets.end(); ++n) {
            (*n)->CalculateXCost();
            (*n)->CalculateYCost();
        }

        double delta1 = 0;
        double delta2 = 0;
        double tx = 0;
        double ty = 0;

        for (ConjugateGradient::variable_iterator b = variables.begin();
                b != variables.end(); ++b) {
            assert(*b);
            CGBlockData* bd = block_data(**b);
            tx = bd->mrx;
            ty = bd->mry;
            CalculateResiduals(*b);
            delta1 += bd->rx * tx;
            delta2 += bd->rx * bd->mrx;
            delta1 += bd->ry * ty;
            delta2 += bd->ry * bd->mry;
        }

        if (delta > 0) {
            beta = (delta2 - delta1) / delta;

            if (beta < 0) {
                beta = 0;
            }
        }

        assert(!isnan(beta));
        assert(!isnan(delta));
        delta = delta2;
    }

    void ConjugateGradient::InitializeData() {
        for (Placement::net_iterator n = place->nets.begin(); n != place->nets.end(); ++n) {
            assert(*n);
            (*n)->CalculateXCost();
            (*n)->CalculateYCost();
        }

        for (ConjugateGradient::variable_iterator b = variables.begin();
                b != variables.end(); ++b) {
            assert(*b);
            block_data(**b)->Initialize();
        }
    }

    double ConjugateGradient::CalculateAlpha() {
        double t1, t2, t3;
        double deno = 0; // d_i^T \cdot f^{\prime\prime}(x_i) \cdot d_i

        assert(!place->nets.empty());

        for (Placement::net_iterator n = place->nets.begin(); n != place->nets.end(); ++n) {
            t1 = t2 = t3 = 0;
            assert(*n);
            for (size_t i = 0; i < (*n)->num_blocks(); ++i) {
                Block* b = (*n)->block(i);
                CGBlockData* bd = block_data(*b);
                t1 += bd->dx * bd->dx;
                t2 += bd->dx;
                t3 += bd->dx * (b->x - (*n)->x);
            }

            t2 /= (*n)->num_blocks();;
            assert((*n)->costx);
            t3 /= (*n)->costx * (*n)->costx;
            deno += (t1 - t2 - t3) / (*n)->costx; // FIXME: square t2 ???
            // deno is equation 4.15 (without alpha)
            // FIXME divide by the cardinality ???

            t1 = t2 = t3 = 0;
            for (size_t i = 0; i < (*n)->num_blocks(); ++i) {
                Block* b = (*n)->block(i);
                CGBlockData* bd = block_data(*b);
                t1 += bd->dy * bd->dy;
                t2 += bd->dy;
                t3 += bd->dy * (b->y - (*n)->y);
            }

            t2 /= (*n)->num_blocks(); // SQUARE IT ????
            assert((*n)->costy);
            t3 /= (*n)->costy * (*n)->costy;
            deno += (t1 - t2 - t3) / (*n)->costy;
        }

        double nume = 0; // r_i^T \cdot r_i

        for (ConjugateGradient::variable_iterator b = variables.begin();
                b != variables.end(); ++b) {
            IOBlock* io = dynamic_cast<IOBlock*>(*b);
            if (io) {
                CGIOData* bd = io_data(*io);
                deno += bd->dx * bd->dx * bd->dxx + bd->dy * bd->dy * bd->dyy + 2 *\
                        bd->dx * bd->dy * bd->dxy;
                nume += bd->dx * bd->rx;
                nume += bd->dy * bd->ry;
            } else {
                CGBlockData* bd = block_data(**b);
                nume += bd->dx * bd->rx;
                nume += bd->dy * bd->ry;
            }
        }

        //printf("Numerator: %f\n", nume);
        //printf("Denominator: %f\n", deno);

        assert(deno != 0.0);
        alpha = nume / deno;
        return alpha; // \alpha_i WHERE IS STAR PLUS ALPHA ???
    }
} // namespace starplacer
