/**
 * @copyright University of Guelph
 */

#include <cstddef>
#include <math.h>
#include <assert.h>
#include <stdio.h>

#include <vector>
#include <list>

#include "Placement.h"
#include "Block.h"
#include "SuccessiveOverRelaxation.h"
#include "CornerPrePlacer.h"

namespace starplacer {

    Block* CornerPrePlacer::GetMaxXBlock() {
        if (placement->blocks.empty()) {
            return NULL;
        }

        Block* max = placement->blocks[0];
        for (int i = 1; i < (int)placement->blocks.size(); ++i) {
            if (max->x < placement->blocks[i]->x) {
                max = placement->blocks[i];
            }
        }

        return max;
    }

    /**
     * This pre-placement algorithm uses SOR's graph traversal
     * to find four loosely connected blocks to place at the corners.
     * This seems to work better then working with all IO Blocks
     *
     * Essentially SORs traversal algorithm gives an ordering
     * of net and blocks which are in decreasing order of connectivity
     * to a given source block.
     *
     * After choosing a corner piece, all blocks are placed at the same position
     * far away from the corner piece. SOR runs and naturally pulls some blocks
     * toward the corner, the block which is least pulled is assigned to the opposite
     * corner.
     *
     * with multiple corners, their orderings for connectivity can be combined (using SOR)
     * into a single sequence and the process repeated from there on until 4 corners are chosen.
     *
     * After choosing the corners and placing them, the rest are put to the centre of the placement
     *
     * SORPlacer then uses these four corners to calculate the new combined sequence from this
     * initial state, and SOR is run with this ordering.
     */
    CornerPrePlacer::CornerPrePlacer(SuccessiveOverRelaxation& sor) {
        placement = sor.placement;

        SuccessiveOverRelaxation::VarSequence src; // source nodes (corners)
        std::vector<SuccessiveOverRelaxation::VarSequence> seqs; // closeness sequences for each corner

        // pick any block (replaced later)
        topleft = placement->blocks[0];
        src.push_back(sor.BlockLookup(topleft));

        seqs.push_back(sor.CalculateSequence(topleft));
        sor.CombineAndOrderSequences(src, seqs); // load the sequence into SOR (sor.sequence)
        SuccessiveOverRelaxation::VarSequence::iterator inode = sor.sequence.begin();
        for (; inode != sor.sequence.end(); ++inode) {
            (*inode)->setPosition(100, 0); // far away
        }

        topleft->x = 0; // far away from the other blocks
        sor.QuickConverge();
        topleft = GetMaxXBlock(); // the block which moved the least

        // use this topleft as the real one (replace the old structures)
        src.front() = sor.BlockLookup(topleft);
        seqs.front() = sor.CalculateSequence(topleft);
        sor.CombineAndOrderSequences(src, seqs);

        inode = sor.sequence.begin();
        for (; inode != sor.sequence.end(); ++inode) {
            (*inode)->setPosition(100, 0);
        }
        topleft->x = 0;

        sor.QuickConverge();
        bottomright = GetMaxXBlock();

        inode = sor.sequence.begin();

        // TODO: Check this, there is another
        // loop ding the same thing below
        // without a ConvergeBlock between
        // the positions of the blocks shouldn't
        // need to be reset twice
        for (; inode != sor.sequence.end(); ++inode) {
            (*inode)->setPosition(100, 0);
        }

        src.push_back(sor.BlockLookup(bottomright));
        seqs.push_back(sor.CalculateSequence(bottomright));
        sor.CombineAndOrderSequences(src, seqs);

        inode = sor.sequence.begin();
        for (; inode != sor.sequence.end(); ++inode) {
            (*inode)->setPosition(100, 0);
        }
        topleft->x = bottomright->x = 0;

        sor.QuickConverge();
        topright = GetMaxXBlock();

        src.push_back(sor.BlockLookup(topright));
        seqs.push_back(sor.CalculateSequence(topright));
        sor.CombineAndOrderSequences(src, seqs);

        inode = sor.sequence.begin();
        for (; inode != sor.sequence.end(); ++inode) {
            (*inode)->setPosition(100, 0);
        }
        topleft->x = topright->x = bottomright->x = 0;

        sor.QuickConverge();
        bottomleft = GetMaxXBlock();

        src.push_back(sor.BlockLookup(bottomleft));
        seqs.push_back(sor.CalculateSequence(bottomleft));
        sor.CombineAndOrderSequences(src, seqs);

        // Position Corners and move the rest to centre
        double x = (placement->arch.nx + 1) / 2.0;
        double y = (placement->arch.ny + 1) / 2.0;

        inode = sor.sequence.begin();
        for (; inode != sor.sequence.end(); ++inode) {
            (*inode)->setPosition(x, y);
        }

        topleft->x = bottomleft->x = 0;
        topright->x = bottomright->x = placement->arch.nx + 1;
        topleft->y = topright->y = 0;
        bottomleft->y = bottomright->y = placement->arch.ny + 1;
    }
}
