/**
 * @copyright University of Guelph
 */

#include <stdio.h>
#include <cstddef>
#include <math.h>
#include <assert.h>

#include <sstream>
#include <string>
#include <iostream>
#include <vector>

#include "vpr_types.h"
#include "yaml.h"
#include "Block.h"

#include "Net.h"

namespace starplacer {
    const double Net::alpha_factor = 1.59;

    Net::Net(int num_pins, int id): id_(id) {
        costx = costy = 1;
        sumx = sumy = sumx2 = sumy2 = 0;
        is_global = false;
        x = y = 0;
        blocks_.reserve(num_pins);
    }

     void Net::print() const {
        printf("id: %d, cost_x: %.2f, cost_y: %.2f, cardinality: %d, ",
                id_, costx, costy, (int)blocks_.size());
    }

    void Net::addBlock(Block* block) {
        assert(block);
        blocks_.push_back(block);
    }

    /**
     * @warn Alpha (1.59) is not applied
     */
    void Net::CalculateXCost() {
        sumx = 0;
        sumx2 = 1; // beta already added

        for (size_t i = 0; i < blocks_.size(); ++i) {
            sumx += blocks_[i]->x;
        }

        assert(blocks_.size() > 0 );
        x = sumx / blocks_.size();

        for (size_t i = 0; i < blocks_.size(); ++i) {
            sumx = blocks_[i]->x - x; // this is not sumx
            sumx2 += sumx * sumx;
            assert(!isnan(sumx2));
        }

        costx = sqrt(sumx2);
        assert(!isnan(costx));
    }

    /**
     * computes from scratch.
     * @warn Alpha or factor (1.59) is not applied
     */
    void Net::CalculateYCost() {
        sumy = 0;
        sumy2 = 1; // beta already added

        for (size_t i = 0; i < blocks_.size(); ++i) {
            sumy += blocks_[i]->y;
            assert(!isnan(sumy));
        }

        assert(blocks_.size() > 0);
        y = sumy / blocks_.size();

        for (size_t i = 0; i < blocks_.size(); ++i) {
            sumy = blocks_[i]->y - y;
            sumy2 += sumy * sumy;
        }

        costy = sqrt(sumy2);
        assert(!isnan(costy));
    }

    void Net::BlockMoved(double oldx, double oldy, double newx, double newy) {
        assert(sumx != 0.0);
        sumx += newx - oldx;
        sumy += newy - oldy;
        sumx2 += (newx - oldx) * (newx + oldx);
        sumy2 += (newy - oldy) * (newy + oldy);
        x = sumx / blocks_.size();
        assert(!isnan(x));
        assert(!isinf(x));
        y = sumy / blocks_.size();

        double tmp = sumx * x;
        assert(sumx2 > tmp);
        costx = sqrt(sumx2 - tmp);
        tmp = sumy * y;
        assert(sumy2 > tmp);
        costy = sqrt(sumy2 - tmp);
    }
} // namespace starplacer

namespace yaml {

    std::vector<int> blockIds(starplacer::Net* net) {
        std::vector<int> ids;
        for (size_t i = 0; i < net->num_blocks(); ++i) {
            ids.push_back(net->block(i)->id());
        }
        return ids;
    }

    std::string yaml(starplacer::Net* net) {
        std::stringstream ss;
        if (net == NULL) {
            return "null";
        }
        ss << "{id: " << net->id();
        ss << ", x: " << net->x;
        ss << ", y: " << net->y;
        ss << ", blocks: " << yaml(blockIds(net));
        ss << ", type: \"Net\"}";
        return ss.str();
    }
}
