/**
 * @copyright University of Guelph
 */

#include <cstddef>
#include <assert.h>
#include <stdio.h>
#include <string.h>

#include <vector>
#include <fstream>

#include "util.h"
#include "vpr_types.h"

#include "IOBlock.h"
#include "Block.h"
#include "Net.h"
#include "Architecture.h"
#include "Netlist.h"
#include "Placement.h"

static int my_round(double num) {
    return num > 0 ? num + .5 : num - .5;
}

namespace starplacer {
   typedef std::vector<Net*>::iterator net_iter;

    Placement::Placement(Netlist& netlist, Architecture& arch): netlist(netlist), arch(arch) {

        blocks.reserve(netlist.num_blocks);
        io_blocks.reserve(netlist.num_blocks);
        clb_blocks.reserve(netlist.num_blocks);
        nets.reserve(netlist.num_nets);

        int io_id = 0;

        for (int i = 0; i < netlist.num_blocks; ++i) {
            if (netlist.block[i].type == CLB) {
                Block* b = new Block(arch.pins_per_clb, i);
                clb_blocks.push_back(b);
                blocks.push_back(b);
            } else {
                IOBlock* b = new IOBlock(i, io_id++);
                io_blocks.push_back(b);
                blocks.push_back(b);
            }
        }

        for (int i = 0; i < netlist.num_nets; ++i) {
            s_net n = netlist.net[i];
            Net* my_net = new Net(n.num_pins, i);
            my_net->is_global = netlist.is_global[i];
            for (int j = 0; j < n.num_pins; ++j) {
                Block* my_block = blocks[n.blocks[j]];
                assert(my_block);
                my_block->addNet(my_net, n.blk_pin[j]);
                my_net->addBlock(my_block);
            }
            nets.push_back(my_net);
        }

        for (Placement::block_iterator ib = blocks.begin(); ib != blocks.end(); ++ib) {
            Block *b = *ib;
            b->CondenseEdges(); // remove NULL's from connected nets.
            if (b->num_nets() == 0) {
                isolated_blocks.push_back(b);
            }
        }
    }

    void Placement::Store() {
        stored_x.resize(netlist.num_blocks);
        stored_y.resize(netlist.num_blocks);
        for (Placement::block_iterator b = blocks.begin(); b != blocks.end(); ++b) {
            stored_x[(*b)->id()] = (*b)->x;
            stored_y[(*b)->id()] = (*b)->y;
        }
    }

    void Placement::Restore() {
        for (Placement::block_iterator b = blocks.begin(); b != blocks.end(); ++b) {
            (*b)->x = stored_x[(*b)->id()];
            (*b)->y = stored_y[(*b)->id()];
        }
    }

    Placement::~Placement() {
        for (block_iterator b = blocks.begin(); b != blocks.end(); ++b) {
           delete *b;
        }

        for (net_iterator n = nets.begin(); n != nets.end(); ++n) {
            delete *n;
        }
    }

    /**
     * Computes the Current Star+ Wirelength cost for the placement.
     * @return the Star+ cost for the entire placement.
     */
    double Placement::StarPlusWirelengthCost() {
        double cost = 0;

        for (net_iterator n = nets.begin(); n != nets.end(); ++n) {
            (*n)->CalculateXCost();
            (*n)->CalculateYCost();
            cost += (*n)->costx + (*n)->costy;
        }

        return Net::alpha_factor * cost;
    }

    /**
     * writes the CLB coordinates (x and y) to file
     * in a tab separated format
     */
    void Placement::WriteToFile(const std::string& filename) const {
    /* Prints out the placement of the circuit.  The architecture and    *
     * netlist files used to generate this placement are recorded in the *
     * file to avoid loading a placement with the wrong support files    *
     * later.                                                            */

     FILE *fp;
     int subblock;

     fp = my_fopen(filename.c_str(), "w", 0);

     fprintf(fp,"Netlist file: %s   Architecture file: %s\n", netlist.filename.c_str(),
        arch.filename.c_str());
     fprintf (fp, "Array size: %d x %d logic blocks\n\n", arch.nx, arch.ny);
     fprintf (fp, "#block name\tx\ty\tsubblk\tblock number\n");
     fprintf (fp, "#----------\t--\t--\t------\t------------\n");

     for (size_t i = 0; i < blocks.size(); i++) {
        fprintf (fp,"%s\t", netlist.block[i].name);
        if (strlen(netlist.block[i].name) < 8)
           fprintf (fp, "\t");
        double block_x = blocks[i]->x;
        double block_y = blocks[i]->y;

        fprintf (fp,"%d\t%d", (int)my_round(block_x), (int)my_round(block_y));
        if (netlist.block[i].type == CLB) {
           fprintf(fp,"\t%d", 0);        /* Sub block number not meaningful. */
        } else {                /* IO block.  Save sub block number. */
           subblock = dynamic_cast<IOBlock*>(blocks[i])->z;
           fprintf(fp,"\t%d", subblock);
        }
        fprintf (fp, "\t#%d\n", (int)i);
     }

     fclose(fp);
    }
} // namespace starplacer


namespace yaml {

    std::string yaml(starplacer::Placement* p) {
        std::stringstream ss;
        if (p == NULL) {
            return "null";
        }
        ss << "{";
        //ss << "Blocks: " << yaml(p->blocks);
        ss << ", ";
        //ss << "Nets: " << yaml(p->nets);
        ss << "}";
        return ss.str();
    }

}
