/**
 * @copyright University of Guelph
 */

#include <cstddef>
#include <math.h>
#include <assert.h>
#include <stdio.h>

#include <list>
#include <vector>

#include "Placement.h"
#include "Block.h"
#include "IOBlock.h"
#include "Net.h"

#include "Shrubbery.h"

namespace starplacer {
   Shrubbery::~Shrubbery() {
        for (HeapNodes::iterator tnode = tnodes_.begin(); tnode != tnodes_.end(); ++tnode) {
            if (*tnode != NULL) {
                delete *tnode;
            }
        }
     }

    Shrubbery::Shrubbery(Placement* placement) {
        place = placement;
        shrub_count = place->netlist.num_p_inputs + place->netlist.num_p_outputs;
        shrubs_.resize(place->netlist.num_blocks, NULL);
        net_shrubs_.resize(place->netlist.num_nets, NULL);
        cycle_memory_.resize(shrub_count);
        cycles_.resize(shrub_count, NULL);
        tnodes_.resize(place->netlist.num_nets, NULL);
        partitions_.resize(shrub_count);

        for (Placement::io_iterator b = place->io_blocks.begin(); b != place->io_blocks.end(); ++b) {
            shrub(**b) = *b;
        }

        for (size_t i = 0; i < shrub_count; ++i) {
            cycles_[i] = &cycle_memory_[i];
        }

        for (Placement::io_iterator s = place->io_blocks.begin(); s != place->io_blocks.end(); ++s) {
            cycle(**s)->clear();
            cycle(**s)->push_back(*s);

            // add each net connected to the io as candidate edges in the heap
            for (size_t i = 0; i < (*s)->num_nets(); ++i) {
                Net* vn = (*s)->nets(i);
                assert(vn);
                if (shrub(*vn) != NULL) {
                    continue;
                }

                tnode(*vn) = new Shrubbery::IOHeap::Node(vn);
                shrub(*vn) = *s;
                tree.Insert(tnode(*vn), vn->num_blocks());
            }
        }

        ExpandShrubs();
        SequenceIO();
        CentreBlocks();
        WrapIOSequence();
    }

    void Shrubbery::ExpandShrubs() {
        size_t meetcount = 0;

        while (tree.size() > 0) {
            Shrubbery::IOHeap::Node* tn = tree.ExtractMin();
            Net* n = tn->id();
            float cost = tn->key();

            for (size_t i = 0; i < n->num_blocks(); ++i) {
                Block* block = n->block(i);

                if (shrub(*block) == NULL) {
                    shrub(*block) = shrub(*n);

                    for (size_t j = 0; j < block->num_nets(); ++j) {
                        Net* vn1 = block->nets(j);
                        assert(vn1);
                        if (shrub(*vn1) != NULL) {
                            continue;
                        }

                        tnode(*vn1) = new Shrubbery::IOHeap::Node(vn1);
                        shrub(*vn1) = shrub(*block);
                        tree.Insert(tnode(*vn1), cost + vn1->num_blocks());
                    }
                } else {
                    SetPartition* b_part = partition(*shrub(*block));
                    SetPartition* n_part = partition(*shrub(*n));

                    if (b_part->isSamePartition(*n_part)) {
                        continue;
                    }

                    MergeCycles(*shrub(*n), *shrub(*block));
                    SetPartition::CombinePartitions(b_part, n_part);
                    if (++meetcount == shrub_count - 1) {
                        break;
                    }
                }
            }
        }
    }

void Shrubbery::SequenceIO() {
    io_sequence.reserve(shrub_count);
    IOBlock* iopad;

    for (Placement::isolated_iterator sio = place->isolated_blocks.begin();
            sio != place->isolated_blocks.end(); ++sio) {
        iopad = dynamic_cast<IOBlock*>(*sio);
        if (iopad != NULL) {
            io_sequence.push_back(iopad);
            cycle(*iopad)->clear();
        }
    }

    for (shrub_iterator s = shrubs_.begin(); s != shrubs_.end(); ++s) {
        assert(*s);

        for (cycle_iterator sio = cycle(**s)->begin(); sio != cycle(**s)->end(); ++sio) {
            io_sequence.push_back(*sio);
        }

        if (io_sequence.size() == shrub_count) {
            break;
        }

        cycle(**s)->clear();
    }
}

void Shrubbery::WrapIOSequence() {
  // place the io pads around the border
        const float step = (place->arch.nx + place->arch.ny + 2) * 2.0f / io_sequence.size();

        int i = 0;
        IOBlock* io;
        std::vector<IOBlock*>::iterator iio = io_sequence.begin();
        float x, y; // current_io_ptr
        y = 0;
        for (x = 0; x <= place->arch.nx + 1; x += step) {
            io = *iio;
            ++iio;
            ++i;
            if (io == NULL) {
                continue;
            }
            io->x = x;
            io->y = 0;
        }

        y = x - place->arch.nx - 1; // wrap the remaining distance around the corner
        x = place->arch.nx + 1;
        for (; y <= place->arch.ny + 1; y += step) {
            io = *iio;
            ++i;
            ++iio;
            if (io == NULL) {
                continue;
            }
            io->x = x;
            io->y = y;
        }

        x = place->arch.nx + 2 - y + place->arch.ny;
        y = place->arch.ny + 1;

        for (; x >= 0; x -= step) {
            io = *iio;
            ++i;
            ++iio;
            if (io == NULL) {
                continue;
            }
            io->x = x;
            io->y = y;
        }

        y = place->arch.ny + 1 + x;
        for (; i < (int)io_sequence.size(); y -= step) {
            io = *iio;
            ++i;
            ++iio;
            if (io == NULL) {
                continue;
            }
            io->x = 0;
            io->y = y;
        }
    }

    void Shrubbery::CentreBlocks() {
        // place all the blocks at the centre.
        const float mid_x = (place->arch.nx + 1) / 2.0f;
        const float mid_y = (place->arch.ny + 1) / 2.0f;

        for (Placement::block_iterator b = place->blocks.begin(); b != place->blocks.end(); ++b) {
            (*b)->x = mid_x;
            (*b)->y = mid_y;
        }
    }

    void Shrubbery::MergeCycles(IOBlock& block, IOBlock& other) {

        // find the position of block in its cycle
        int index = -1;
        for (cycle_iterator io = cycle(block)->begin(); io != cycle(block)->end(); ++io) {
            index += 1;
            if ((*io)->id() == block.id()) {
                break;
            }
        }

        index = index * 2 + 1;

        // find the position of other in its cycle
        int otherindex = -1;
        for (cycle_iterator io = cycle(other)->begin(); io != cycle(other)->end(); ++io) {
            otherindex += 1;
            if ((*io)->id() == other.id()) {
                break;
            }
        }

        otherindex = otherindex * 2 + 1;

        if (index < (int)cycle(block)->size()) {
            if (otherindex < (int)cycle(other)->size()) {
                cycle(other)->reverse();
            }

            cycle(other)->insert(cycle(other)->end(), cycle(block)->begin(), cycle(block)->end());
            cycle_iterator end = cycle(block)->end(); // end may change in the process of iterating so save it here

            for (cycle_iterator io = cycle(block)->begin(); io != end; ++io) {
                cycle(**io) = cycle(other);
            }
        } else {
            if (otherindex > (int)cycle(other)->size()) {
                cycle(other)->reverse();
            }

            cycle(block)->insert(cycle(block)->end(), cycle(other)->begin(), cycle(other)->end());
            cycle_iterator end = cycle(other)->end();

            for (cycle_iterator io = cycle(other)->begin(); io != end; ++io) {
                cycle(**io) = cycle(block);
            }
        }
    }
} // namespace starplacer
