/**
 * @copyright University of Guelph
 */

#include <cstddef>
#include <math.h>
#include <assert.h>
#include <stdio.h>

#include <algorithm>
#include <limits>
#include <vector>
#include <list>

#include "Block.h"
#include "IOBlock.h"
#include "DoubleEndedPriorityQueue.h"
#include "Placement.h"

#include "SimpleBipartition.h"

namespace starplacer {

    void SimpleBipartition::Initialize(std::vector<Block*> blocks,
            std::vector<IOBlock*> io_blocks) {
        clbs = blocks;
        ios = io_blocks;
        if (sub1) {
            delete sub1;
            sub1 = NULL;
        }

        if (sub2) {
            delete sub2;
            sub2 = NULL;
        }

        for (std::vector<Block*>::iterator it = clbs.begin(); it != clbs.end(); ++it) {
            int index = (*it)->id();
            if (tNodeX[index]) {
                delete tNodeX[index];
            }
            if (tNodeY[index]) {
                delete tNodeY[index];
            }
            tNodeY[index] = new HeapNode(*it);
            tNodeX[index] = new HeapNode(*it);
        }

        for (std::vector<IOBlock*>::iterator it = ios.begin(); it != ios.end(); ++it) {
            int index = (*it)->id();
            if (tNodeX[index]) {
                delete tNodeX[index];
            }
            if (tNodeY[index]) {
                delete tNodeY[index];
            }
            tNodeX[index] = new HeapNode(*it);
            tNodeY[index] = new HeapNode(*it);
        }
    }

    SimpleBipartition* SimpleBipartition::root() {
        return root_ ? root_ : this;
    }

    SimpleBipartition::SimpleBipartition(Placement* placement, double density) {
        root_ = NULL;
        this->placement = placement;
        int num_blocks = placement->netlist.num_blocks;
        tNodeX.resize(num_blocks, NULL);
        tNodeY.resize(num_blocks, NULL);
        InitRegion(0, placement->arch.ny + 1, 0, placement->arch.nx + 1, density);
    }

    /**
     * This algorithm places all blocks at the centre of the tile.
     * It ignores the possibility of having more than one allowable
     * CLB in a single grid position (the algorithm ignores the z dimension)
     * however for most placements it is not possible to ignore the z dimension
     * of the IO.
     */
    void SimpleBipartition::SpreadIO() {
        std::vector<int> occ_count;
        int index;
        int x, y;
        Placement* p = root()->placement;

        // indexed by left, right, top, then bottom
        occ_count.resize(2 * (p->arch.nx + p->arch.ny), 0);
        for (Placement::io_iterator io = p->io_blocks.begin(); io != p->io_blocks.end(); ++io) {
            x = (*io)->x;
            y = (*io)->y;

            if (x > p->arch.nx) {
                index = p->arch.ny + y;
            } else if (x == 0) {
                index = y;
            } else if (y > p->arch.ny) {
                index = 2 * p->arch.ny + p->arch.nx + x;
            } else {
                assert(y == 0);
                index = 2 *p-> arch.ny + x;
            }

            (*io)->z = occ_count[index]++;
        }
    }

    bool IOXLessThan(const IOBlock* a, const IOBlock* b) {
        return a->x < b->x;
    }

    bool IOYLessThan(const IOBlock* a, const IOBlock* b) {
        return a->y < b->y;
    }

    SimpleBipartition::~SimpleBipartition() {
        if (root_ == this) {
            for (size_t i = 0; i < tNodeX.size(); ++i) {
                delete tNodeX[i];
                delete tNodeY[i];
            }
            tNodeX.resize(0);
            tNodeY.resize(0);
        }

        if (sub1) delete sub1;
        if (sub2) delete sub2;
    }


    // Defines a rectangular area of the FPGA, according to the global architecture
    // where top is the lower y coordinate, and l is the lower x coordinate
    // the maximum number of clbs and ios (according to the area and the architecture are set here)
    // the minimum number is set relative to the another region in initSub
    // if the region is a column or row vector of ios only, then the region will shrink to exclude the corner pieces.
    // this makes sure that regions are always rectangular by allowing them to include a corner piece and removing
    // them when the result would still be a rectangle.

    SimpleBipartition::SimpleBipartition(SimpleBipartition* root, int t, int b, int l, int r, double density) {
        root_ = root;
        InitRegion(t, b, l, r, density);
    }

    void SimpleBipartition::InitRegion(int t, int b, int l, int r, double density) {
        assert(b >= t);
        assert(r >= l);

        sub1 = NULL;
        sub2 = NULL;
        top = t;
        bottom = b;
        left = l;
        right = r;
        density_ = density;

        if (l == 0) {
            ++l;
        }

        if (r > root()->placement->arch.nx) {
            --r;
        }

        if (t == 0) {
            ++t;
        }

        if (b > root()->placement->arch.ny) {
            --b;
        }

        int width = r - l + 1;
        int height = b - t + 1;

        maxclbs = (int)ceil(width * height * density);

        if (width == 0) {
            if (top == 0) {
                ++top;
            }

            if (bottom > root()->placement->arch.ny) {
                --bottom;
            }
        }

        if (height == 0) {
            if (left == 0) {
                ++left;
            }

            if (right > root()->placement->arch.nx) {
                --right;
            }
        }

        x = (left + right) / 2.0;
        y = (top + bottom) / 2.0;

        maxios = 0;

        if (top == 0) {
            maxios += width;
        }

        if (bottom > root()->placement->arch.ny) {
            maxios += width;
        }

        if (left == 0) {
            maxios += height;
        }

        if (right > root()->placement->arch.nx) {
            maxios += height;
        }

        maxios *= root()->placement->arch.io_rat; //

        clbs.clear(); //resize(maxclbs);
    }

    // non recursive
    void SimpleBipartition::partitionIO() {
        int i;

        if (right - left > bottom - top) {
            std::sort(ios.begin(), ios.end(), IOXLessThan);
            i = sub1->minios;
            std::vector<IOBlock*>::iterator Iio = ios.begin();
            std::advance(Iio, i);

            while (i < sub1->maxios && (*Iio)->x < x) {
                ++i;
                ++Iio;
            }
        } else {
            std::sort(ios.begin(), ios.end(), IOYLessThan);
            i = sub1->minios;
            std::vector<IOBlock*>::iterator Iio = ios.begin();
            std::advance(Iio, i);

            while (i < sub1->maxios && (*Iio)->y < y) {
                ++i;
                ++Iio;
            }
        }

        sub1->ios.clear();
        sub1->ios.reserve(i);
        std::vector<IOBlock*>::iterator Iio = ios.begin();

        for (int j = 0; j < i; ++j, ++Iio) {
            sub1->ios.push_back(*Iio);
        }

        sub2->ios.clear(); //resize(ios.size() - sub1->ios.size());
        sub2->ios.reserve(ios.size() - sub1->ios.size());
        for (int j = i; j < (int)ios.size(); ++j, ++Iio) {
            sub2->ios.push_back(*Iio);
        }
    }

    // recursive
    void SimpleBipartition::Partition() {
        if (right == left && bottom == top) {
            if (clbs.size() > 0 || ios.size() > 0) {
                position();
            }
            return;
        }

        if (right - left > bottom - top) {
            partitionX();
        } else {
            partitionY();
        }

        partitionIO();
        sub1->Partition();
        sub2->Partition();
    }

    void SimpleBipartition::partitionX() {
        if (left == right) {
            return;
        }

        int middle = (left + right) / 2;
        sub1 = new SimpleBipartition(root(), top, bottom, left, middle, density_);
        sub2 = new SimpleBipartition(root(), top, bottom, middle + 1, right, density_);
        initSub();

        x = middle + 0.5;

        typedef std::vector<Block*>::iterator Iblock;
        for (Iblock b = clbs.begin(); b != clbs.end(); ++b) {
            if ((*b)->x <= x) {
                sub1->clbs.push_back(*b);
            } else {
                sub2->clbs.push_back(*b);
            }
        }

        moveSubCLBx(sub1, sub2);
        moveSubCLBx(sub2, sub1);
    }

    void SimpleBipartition::moveSubCLBx(SimpleBipartition* from, SimpleBipartition* to) {

        int numToMove = from->clbs.size() - from->maxclbs;

        if (numToMove <= 0) {
            return;
        }

        std::vector<Block*>temp; //size(from->maxclbs);
        DoubleEndedPriorityQueue<Block*> tree;
        DoubleEndedPriorityQueue<Block*>::Node* maxNode;

        double k1;
        double maxk = std::numeric_limits<double>::max();

        typedef std::vector<Block*>::iterator  Iblock;

        for (Iblock b = from->clbs.begin(); b != from->clbs.end(); ++b) {
            k1 = from->x < x ? -(*b)->x : (*b)->x;

            if (k1 >= maxk) {
                temp.push_back(*b);
                continue;
            }

            tree.Insert(root()->tNodeX[(*b)->id()], k1);
            if (tree.size() <= numToMove) {
                continue;
            }

            maxNode = tree.ExtractMax();
            maxk = maxNode->key();
            temp.push_back(maxNode->id());
        }

        from->clbs.clear();
        from->clbs.insert(from->clbs.end(), temp.begin(), temp.end());

        while (tree.size() > 0) {
            to->clbs.push_back(tree.ExtractMax()->id());
        }

        x = maxk > 0 ? maxk : -maxk;
    }

    /*
    void SimpleBipartition::placeIsolated(CLBGrid& matrix, std::vector<Block*>& isolated) {
        int pos = 0, xy = 0, i = 0, j = 0;

        typedef std::vector<Block*>::iterator Iblock;
        for (Iblock b = isolated.begin(); b != isolated.end(); ++b) {
            if (dynamic_cast<IOBlock*>(*b)) {
               matrix.addToAvailableIO(*b, &pos, &xy);
            } else {
                matrix.addToAvailableCLB(*b, &i, &j);
            }
        }
    }
    */

    void SimpleBipartition::moveSubCLBy(SimpleBipartition* from, SimpleBipartition* to) {
        int numToMove = from->clbs.size() - from->maxclbs;

        if (numToMove <= 0) {
            return;
        }

        std::vector<Block*>temp;
        temp.reserve(from->maxclbs);
        DoubleEndedPriorityQueue<Block*> tree;
        DoubleEndedPriorityQueue<Block*>::Node* maxNode;
        double k1;
        double maxk = std::numeric_limits<double>::max();

        typedef std::vector<Block*>::iterator  Iblock;

        for (Iblock b = from->clbs.begin(); b != from->clbs.end(); ++b) {
            k1 = from->y < y ? -(*b)->y : (*b)->y;

            if (k1 >= maxk) {
                temp.push_back(*b);
                continue;
            }

            tree.Insert(root()->tNodeY[(*b)->id()], k1);
            if (tree.size() <= numToMove) {
                continue;
            }

            maxNode = tree.ExtractMax();
            maxk = maxNode->key();
            temp.push_back(maxNode->id());
        }

        from->clbs.clear();
        from->clbs.insert(from->clbs.end(), temp.begin(), temp.end());

        while (tree.size() > 0) {
            to->clbs.push_back(tree.ExtractMax()->id());
        }

        y = maxk > 0 ? maxk : -maxk;
    }

    void SimpleBipartition::partitionY() {
        if (top == bottom) {
            return;
        }

        const int middle = (top + bottom) / 2;

        sub1 = new SimpleBipartition(root(), top, middle, left, right, density_);
        sub2 = new SimpleBipartition(root(), middle + 1, bottom, left, right, density_);
        initSub();

        y = middle + 0.5;

        typedef std::vector<Block*>::iterator Iblock;
        for (Iblock b = clbs.begin(); b != clbs.end(); ++b) {
            if ((*b)->y <= y) {
                sub1->clbs.push_back(*b);
            } else {
                sub2->clbs.push_back(*b);
            }
        }

        moveSubCLBy(sub1, sub2);
        moveSubCLBy(sub2, sub1);
    }

    void SimpleBipartition::initSub() {
        if (sub1->maxclbs > (int)clbs.size()) {
            sub1->maxclbs = clbs.size();
        }

        if (sub1->maxios > (int)ios.size()) {
            sub1->maxios = ios.size();
        }

        if (sub2->maxclbs > (int)clbs.size()) {
            sub2->maxclbs = clbs.size();
        }

        if (sub2->maxios > (int)ios.size()) {
            sub2->maxios = ios.size();
        }

        sub1->minclbs = (int)clbs.size() - sub2->maxclbs;
        sub1->minios = (int)ios.size() - sub2->maxios;
        sub2->minclbs = (int)clbs.size() - sub1->maxclbs;
        sub2->minios = (int)ios.size() - sub1->maxios;
    }

    void SimpleBipartition::position() {
        typedef std::vector<Block*>::iterator Iblock;

        for (Iblock b = clbs.begin(); b != clbs.end(); ++b) {
            (*b)->x = this->x;
            (*b)->y = y;
        }

        typedef std::vector<IOBlock*>::iterator Iio;
        for (Iio b = ios.begin(); b != ios.end(); ++b) {
            (*b)->x = this->x;
            (*b)->y = y;
        }
    }
}
