/**
 * @copyright University of Guelph
 */

#include <cstddef>
#include <math.h>
#include <assert.h>
#include <stdio.h>

#include <list>
#include <vector>
#include <fstream>
#include <iostream>
#include <sstream>
#include <string>

#include "yaml.h"
#include "Block.h"
#include "IOBlock.h"
#include "Net.h"
#include "Placement.h"

#include "SuccessiveOverRelaxation.h"

namespace starplacer {
    double SuccessiveOverRelaxation::IONode::IOWeight = 0.0;

    SuccessiveOverRelaxation::BlockNode* SuccessiveOverRelaxation::BlockLookup(Block* b) {
        return blocks_[b->id()];
    }

    SuccessiveOverRelaxation::NetNode* SuccessiveOverRelaxation::NetLookup(Net* net) {
        return nets_[net->id()];
    }

    SuccessiveOverRelaxation::SuccessiveOverRelaxation(Placement* placement) {
        assert(placement);
        this->placement = placement;
        blocks_.resize(placement->netlist.num_blocks, NULL);
        nets_.resize(placement->netlist.num_nets, NULL);

        for (Placement::net_iterator n = placement->nets.begin();
                n != placement->nets.end(); ++n) {
            NetNode* nn = new NetNode();
            nn->net = *n;
            nn->CalculateWeight();
            nets_[(*n)->id()] = nn;
        }

        for (Placement::io_iterator b = placement->io_blocks.begin();
                b != placement->io_blocks.end(); ++b) {
            IONode* bn = new IONode();
            bn->block = *b;
            bn->CalculateWeight();
            bn->CalculateOmega(blocks_);
            blocks_[(*b)->id()] = bn;
        }

        for (Placement::block_iterator b = placement->clb_blocks.begin();
                b != placement->clb_blocks.end(); ++b) {
            BlockNode* bn = new BlockNode();
            bn->block = *b;
            bn->CalculateWeight();
            bn->CalculateOmega(blocks_);
            blocks_[(*b)->id()] = bn;
        }

        for (SuccessiveOverRelaxation::netnode_iterator ni = nets_.begin(); ni != nets_.end(); ++ni) {
            (*ni)->CalculateOmega(blocks_);
        }
    }

    SuccessiveOverRelaxation::~SuccessiveOverRelaxation() {
        for (std::vector<BlockNode*>::iterator b = blocks_.begin(); b != blocks_.end(); ++b) {
            delete *b;
        }

        for (std::vector<NetNode*>::iterator n = nets_.begin(); n != nets_.end(); ++n) {
            delete *n;
        }
    }

    void SuccessiveOverRelaxation::NetNode::dump() {
        printf("%d\n", net->id());
        //printf("NetNode: net:%d weight:%f LPindex:%d omega: %f\n",
        //        net->id(), weight, LPindex, omega);
    }

    void SuccessiveOverRelaxation::IONode::dump() {
        printf("%d\n", block->id());
        //printf("IOBlockNode: block:%d weight:%f LPindex:%d omega: %f\n",
        //        block->id(), weight, LPindex, omega);
    }

    void SuccessiveOverRelaxation::BlockNode::dump() {
        printf("%d\n", block->id());
        //printf("BlockNode: block:%d weight:%f LPindex:%d omega: %f\n",
        //        block->id(), weight, LPindex, omega);
    }

    void SuccessiveOverRelaxation::dump_data(std::string filename) {
        FILE* fp = fopen(filename.c_str(), "w");

        fprintf(fp, "id,\t\tx,\t\ty,\t\tweight,\t\tLPindex,\t\tomega\n");

        for (BlockNodes::iterator ib = blocks_.begin(); ib != blocks_.end(); ++ib) {
            BlockNode* bnode = *ib;
            fprintf(fp, "%d,\t\t%f,\t\t%f,\t\t%f,\t\t%d,\t\t%f\n",
                    bnode->block->id(), bnode->block->x,
                    bnode->block->y, bnode->weight,
                    bnode->LPindex, bnode->omega);
        }

        for (NetNodes::iterator in = nets_.begin(); in != nets_.end(); ++in) {
            NetNode* nnode = *in;
            fprintf(fp, "%d,\t\t%f,\t\t%f,\t\t%f,\t\t%d,\t\t%f\n",
                    nnode->net->id(), nnode->net->x,
                    nnode->net->y, nnode->weight,
                    nnode->LPindex, nnode->omega);
        }

        fclose(fp);
    }

    /**
     * @param[in] src: A block in the placement given at initialization of
     * `Successiveoverrelaxation(Placement* p)`
     *
     * @return a sequence of nets and blocks in order of "closeness" to
     * the given source node.
     *
     * @note: this is only used by the CornerPrePlacer,
     * Given a source node, this method calculates the sequence to evaluate the
     * variables in by using Djikstra's algorithm (on a hyper-graph) and
     * using closeness as the weight function
     *
     * @note the return does not include the source, and it does include
     * nets
     */
    SuccessiveOverRelaxation::VarSequence SuccessiveOverRelaxation::CalculateSequence(Block* src) {
        DoubleEndedPriorityQueue<SORNode*> tree;

        // initialize the heap
        for (BlockNodes::iterator b = blocks_.begin(); b != blocks_.end(); ++b) {
            if ((*b)->weight > 0.0f) { // not isolated
                tree.Insert((*b)->heap_node, 0);
            }
        }

        for (SuccessiveOverRelaxation::netnode_iterator n = nets_.begin(); n != nets_.end(); ++n) {
            tree.Insert((*n)->heap_node, 0.0f);
        }

        // remove the source from the heap
        BlockNode* src_node = blocks_[src->id()];
        tree.Remove(src_node->heap_node);

        // update the nets to reflect their new cost from the source node
        for (int i = 0; i < src_node->num_nodes(); ++i) {
            SORNode* n = src_node->node(i, blocks_, nets_);
            // TODO: it seems like re-checking the weight would be faster than testing the heap
            if (!tree.Contains(n->heap_node)) { // ignore isolated, and source
                continue;
            }
            assert(n->weight > 0);
            // its being used as a min-heap so we subtract the connectivity
            tree.set_key(n->heap_node, n->heap_node->key() - n->EdgeWeight(n) / n->weight);
        }

        // Djikstra's algorithm with edges weighed by "connectivity"
        VarSequence seq; // the order in which the nodes were visited
        //seq.reserve(tree.size());
        //printf("--- begin sequence\n");
        while (tree.size() > 0) {
            SORNode* node = tree.ExtractMin()->id();
            //printf("--- begin update\n");
            for (int i = 0; i < node->num_nodes(); ++i) {
                SORNode* n = node->node(i, blocks_, nets_);
                if (!tree.Contains(n->heap_node)) {
                    continue;
                }
                assert(node->weight > 0);
                tree.set_key(n->heap_node, n->heap_node->key() - node->EdgeWeight(n) / node->weight);
                //n->dump();
            }
            //printf("--- end update\n");
            //node->dump();
            seq.push_back(node);
        }
        //printf("--- end sequence\n");
        return seq;
    }

    /**
     * @param[in] src: a list of the source nodes
     * @param[in] seqs: a list of sequences of blocks and nets
     * For each `src[i]`, the corresponding sequence `seqs[i]` is
     * a sequence of blocks and nets in order of "closeness" to the `src[i]`
     * The result is a sequence where each block appears at most
     * once and is given the LPindex which is the earliest index in
     * any of the sequences that it appears at.
     * The result is stored in the internal public member sequence.
     *
     * @note the source and sequences are produced by
     * SuccessiveOverRelaxation::CalculateSequence(src) -> seq
     */
    void SuccessiveOverRelaxation::CombineAndOrderSequences(VarSequence& src, std::vector<VarSequence>& seqs) {
        // clear the current sequence and invalidate all LP indices
        sequence.clear();
        for (SuccessiveOverRelaxation::blocknode_iterator b = blocks_.begin();
                b != blocks_.end(); ++b) {
            (*b)->LPindex = -1;
        }

        for (SuccessiveOverRelaxation::netnode_iterator n = nets_.begin();
                n != nets_.end(); ++n) {
            (*n)->LPindex = -1;
        }

        // sources are shown in other lists
        // but not the ones calculated with them as the source
        // and they are of optimal closeness to a source (themselves) -> index = 0
        for (VarSequence::iterator n = src.begin();
                n != src.end(); ++n) {
            (*n)->LPindex = 0;
        }

        size_t num_nodes = placement->blocks.size() + placement->nets.size();
        unsigned int max_count = num_nodes - src.size() - placement->netlist.num_globals
            - placement->isolated_blocks.size();

        // initialize the iterators, (lists don't have random access)
        std::vector<VarSequence::iterator> current;
        for (size_t j = 0; j < seqs.size(); ++j) {
            current.push_back(seqs[j].begin());
        }

        unsigned int count = 0;

        for (size_t i = 0; i < num_nodes; ++i) {
            for (size_t j = 0; j < seqs.size(); ++j) {
                SORNode* node = *current[j];
                ++current[j];
                if (node->LPindex >= 0) { // this node was spotted earlier in another sequence
                    continue;
                }
                node->LPindex = i + 1; // +1 since source are given 0
                sequence.push_back(node);

                //if (current[j] == seqs[j].end()) return;
                if (++count == max_count) { // optimization, found all there could be.
                    return;
                }
            }
        }
    }

    /**
     * @precondition @ref CombineAndOrderSequences must have been called
     * before this to give the order to use in evaluating the blocks and nets.
     * @note this is called after the Cornerpreplacer, and SORPlacer have
     * called Combineandordersequences
     *
     * The `Go` method of the algorithm, after this is called the block
     * positions and net positions will be updated, the result modifies the
     * placement blocks and nets directly, after this is done
     * a legalization method usually follows before calling this again since it will generate
     * placements which have a significant amount of overlap
     */
    void SuccessiveOverRelaxation::QuickConverge() {
        float ome = 1 - (1.0 / sqrt(sequence.size()));

        for (SuccessiveOverRelaxation::VarSequence::iterator n = sequence.begin();
                n != sequence.end(); ++n) {
            (*n)->CalculatePosition(ome, true, placement, blocks_, nets_);
        }

        for (SuccessiveOverRelaxation::VarSequence::iterator n = sequence.begin();
                n != sequence.end(); ++n) {
            (*n)->CalculatePosition((*n)->omega, false, placement, blocks_, nets_);
        }

        for (SuccessiveOverRelaxation::VarSequence::iterator n = sequence.begin(); n != sequence.end(); ++n) {
            (*n)->CalculatePosition((*n)->omega, false, placement, blocks_, nets_);
        }

        int rounds = 2;
        for (int i = 0; i < rounds; ++i) {
            for (VarSequence::iterator n = sequence.begin(); n != sequence.end(); ++n) {
                (*n)->CalculatePosition(1, false, placement, blocks_, nets_);
            }
        }
    }

    /**
     * The sequence will contain nets after being combined since these are also considered
     * as Nodes and they need to be used in the sequence to ease the transition between
     * stale block positions and new block positions. @see NetNode::CalculatePosition.
     *
     * this method is called by SORPlacer to remove the nets so that it can call
     * ConvergeBlock on the blocks.
     */
    void SuccessiveOverRelaxation::RemoveNets() {
        int count = (int)sequence.size();
        for (int i = 0; i < count; ++i) {
            SORNode* node = sequence.front();
            sequence.pop_front();
            assert(node);
            if (dynamic_cast<NetNode*>(node) == NULL) {
                sequence.push_back(node);
            }
        }
    }

    void SuccessiveOverRelaxation::UpdateNets() {
        for (SuccessiveOverRelaxation::netnode_iterator n = nets_.begin();
                n != nets_.end(); ++n) {
                (*n)->CalculatePosition(1, false, placement, blocks_, nets_);
        }
    }

    /**
     *
     */
    void SuccessiveOverRelaxation::ConvergeBlock(int my_rounds) {
        for (int i = 0; i < my_rounds; ++i) {
            for (SuccessiveOverRelaxation::VarSequence::iterator n = sequence.begin();
                    n != sequence.end(); ++n) {
                BlockNode* bn = dynamic_cast<BlockNode*>(*n);
                assert(bn);
                if (bn) {
                    bn->CalculatePosition(placement);
                }
            }
        }
    }

   int SuccessiveOverRelaxation::NetNode::num_nodes() {
       return net->num_blocks();
   }

   SuccessiveOverRelaxation::SORNode* SuccessiveOverRelaxation::NetNode::node(int i, BlockNodes& block_nodes, NetNodes& net_nodes) {
       return block_nodes[net->block(i)->id()];
   }

   SuccessiveOverRelaxation::SORNode* SuccessiveOverRelaxation::BlockNode::node(int i, BlockNodes& block_nodes, NetNodes& net_nodes) {
       return net_nodes[block->nets(i)->id()];
   }

   int SuccessiveOverRelaxation::BlockNode::num_nodes() {
       return block->num_nets();
   }

   /**
    * the weight of an edge to an io block
    * which is the sum of the reciprocal of the cardinalities of the nets it is connected to.
    * i.e. a block which is connected to few/small nets has a large weight associated with it.
    */
   void SuccessiveOverRelaxation::IONode::CalculateWeight() {
        double w = IONode::IOWeight;
        for (size_t i = 0; i < block->num_nets(); ++i) {
            Net* net = block->nets(i);
            w += 1 / sqrt(net->num_blocks());
        }
        //printf("- {t: block, id: %d, weight: %f}\n", block->id(), w);
        weight = w;
   }

   /**
    * the weight of an edge to a block
    * which is the sum of the reciprocal of the cardinalities of the nets it is connected to.
    * i.e. a block which is connected to few/small nets has a large weight associated with it.
    */
   void SuccessiveOverRelaxation::BlockNode::CalculateWeight() {
        double w = 0;
        for (size_t i = 0; i < block->num_nets(); ++i) {
            Net* n = block->nets(i);
            if (n->is_global) {
                continue;
            }
            w += 1 / sqrt(n->num_blocks());
        }
        //printf("- {t: block, id: %d, weight: %f}\n", block->id(), w);
        weight = w;
    }

    void SuccessiveOverRelaxation::BlockNode::CalculateOmega(BlockNodes& block_nodes) {
        if (weight <= 0) {
            return;
        }

        double ome = 0;

        for (size_t i = 0; i < block->num_nets(); ++i) {
            Net* n = block->nets(i);
            if (n->is_global) {
                continue;
            }

            ome += pow(n->num_blocks(), -1.5);
        }

        ome /= weight;
        ome = ome < 0.5 ? 1 / (1 - ome) : 2;
        omega = ome;
        //printf("- {t: block, id: %d, omega: %f}\n", block->id(), omega);
    }

    void SuccessiveOverRelaxation::NetNode::CalculateOmega(BlockNodes& block_nodes) {
        double ome = 0;

        for (size_t i = 0; i < net->num_blocks(); i++) {
            BlockNode* bn = block_nodes[net->block(i)->id()];
            ome += 1 / bn->weight;
        }

        ome /= pow(net->num_blocks(), 1.5);
        ome = ome < 0.5 ? 1 / (1 - ome) : 2;
        omega = ome;
        //printf("- {t: net, id: %d, omega: %f}\n", net->id(), omega);
    }

    /**
     * Moves the io block toward the centre, of the centres of the nets connected to it.
     *
     * @param ome the distance moved is controlled by ome (omega in the paper) a value between (0 1]
     * @param quick ignore nets who have not yet been updated this iteration
     * @param Placement needed for architecture dimensions
     *
     * @note IOWeight is always 0.0f, I'm not sure what its purpose was.
     */
    void SuccessiveOverRelaxation::IONode::CalculatePosition(double ome, bool quick, Placement* placement,
            BlockNodes& block_nodes, NetNodes& net_nodes) {
        double wx = abs(block->x / (placement->arch.nx + 1) - 0.5) * IONode::IOWeight;
        double wy = abs(block->y / (placement->arch.ny + 1) - 0.5) * IONode::IOWeight;
        double newx = block->x * wx;
        double newy = block->y * wy;

        for (size_t i = 0; i < block->num_nets(); ++i) {
            Net *n = block->nets(i);
            assert(n);
            NetNode* nn = net_nodes[n->id()];
            if (quick && nn->LPindex > LPindex) {
                continue;
            }

            wx++;
            wy++;
            newx += n->x;
            newy += n->y;
        }

        if (wx == 0 || wy == 0) {
            return;
        }

        newx /= wx;
        newy /= wy;

        block->x += (newx - block->x) * ome;
        block->y += (newy - block->y) * ome;

        assert(!isnan(block->x));
        assert(!isnan(block->y));
    }

    /**
     *
     * Moves the block toward the weighted (by cardinality) average of the
     * centres of the nets which it is connected to.
     * @param ome the amount (0 1] which to move toward the new position
     * @param quick ignore nets which have not yet been updated this iteration
     */
    void SuccessiveOverRelaxation::BlockNode::CalculatePosition(double ome, bool quick, Placement* placement, BlockNodes& block_nodes, NetNodes& net_nodes) {
        double wx, wy, newx, newy;
        double tmp;

        wx = wy = newx = newy = 0;

        for (size_t i = 0; i < block->num_nets(); ++i) {
            Net* n = block->nets(i);
            NetNode* nn = net_nodes[n->id()];
            if (n->is_global || (quick && nn->LPindex > LPindex)) {
                continue;
            }

            wx += tmp = Net::alpha_factor / n->costx;
            assert(!isnan(wx) && !isinf(wx));
            newx += n->x * tmp;
            wy += tmp = Net::alpha_factor / n->costy;
            assert(!isnan(wy) && !isinf(wy));
            newy += n->y * tmp;
        }

        assert(wx > 0);
        assert(wy > 0);

        newx /= wx;
        newy /= wy;

        assert(!isnan(block->x));
        assert(!isnan(block->y));

        block->x += (newx - block->x) * ome;
        block->y += (newy - block->y) * ome;
    }

    /**
     * Updates the net's centre point by moving it toward the
     * centre of the blocks connected to it.
     *
     * @param ome how much to move toward the new position (0 1]
     * @param quick ignore nets not yet updated this iteration
     */
    void SuccessiveOverRelaxation::NetNode::CalculatePosition(double ome, bool quick, Placement* placement, BlockNodes& block_nodes, NetNodes& net_nodes) {
        double newx, newy;
        net->sumx = net->sumy = 0;
        net->sumx2 = net->sumy2 = 1;

        int w = 0;

        for (size_t i = 0; i < net->num_blocks(); ++i) {
            Block* b = net->block(i);
            BlockNode* bb = block_nodes[b->id()];

            if (quick && bb->LPindex > LPindex) {
                continue;
            }

            assert(!isnan(b->x));
            net->sumx += b->x;
            assert(!isinf(net->sumx));

            net->sumx2 += b->x * b->x;
            assert(!isinf(net->sumx2));

            net->sumy += b->y;
            assert(!isinf(net->sumy));
            net->sumy2 += b->y * b->y;
            assert(!isinf(net->sumy2));

            w++;
        }

        if (w == 0) {
            return;
        }

        newx = net->sumx / w;
        newy = net->sumy / w;

        double tmp = net->sumx * newx;
        if (net->sumx2 <= tmp) {
            net->sumx2 = tmp;
            net->costx = 0;
        } else  {
            net->costx = sqrt(net->sumx2 - tmp);
        }

        tmp = net->sumy * newy;
        if (net->sumy2 <= tmp) {
            net->sumy2 = tmp;
            net->costy = 0;
        } else {
            net->costy = sqrt(net->sumy2 - tmp);
        }

        assert(!isnan(net->costx) && !isnan(net->costy));

        net->x += (newx - net->x) * ome;
        net->y += (newy - net->y) * ome;
    }

    /**
     * Updates the position of the IO to be the middle of the centres of
     * the connected nets, and triggers an update on the nets.
     *
     * @note the IOWeight pulls IOs that are connected to fewer nets
     * toward the centre of chip (far from the corners).
     * this weight is always set to 0 (not used).
     */
    void SuccessiveOverRelaxation::IONode::CalculatePosition(Placement* placement) {
        double wx = abs(block->x / (placement->arch.nx + 1) - 0.5) * IONode::IOWeight;
        double wy = abs(block->y / (placement->arch.ny + 1) - 0.5) * IONode::IOWeight;
        double newx = block->x * wx;
        double newy = block->y * wy;

        for (size_t i = 0; i < block->num_nets(); ++i) {
            Net* n = block->nets(i);
            wx++;
            wy++;
            newx += n->x;
            newy += n->y;
        }

        if (wx == 0 || wy == 0) {
            return;
        }

        newx /= wx;
        newy /= wy;

        for (size_t i = 0; i < block->num_nets(); ++i) {
            block->nets(i)->BlockMoved(block->x, block->y, newx, newy);
        }

        block->x = newx;
        block->y = newy;

        assert(!isnan(block->x));
        assert(!isnan(block->y));
    }

    double SuccessiveOverRelaxation::BlockNode::EdgeWeight(SORNode* node) {
        NetNode* netnode = dynamic_cast<NetNode*>(node);
        return 1 / sqrt(netnode->net->num_blocks());
    }

    double SuccessiveOverRelaxation::NetNode::EdgeWeight(SORNode* node) {
        //assert(NULL == dynamic_cast<NetNode*>(node)); // nets dont connect to other nets.
        return 1;
    }

    void SuccessiveOverRelaxation::NetNode::CalculateWeight() {
        weight = net->num_blocks();
        //printf("- {t: net, id: %d, weight: %f}\n", net->id(), weight);
    }

    /**
     * This method modifies the position given by adjusting it to the timing
     * measurements.
     * TODO: add this functionality, put timing_tradeoff
     * somewhere accessible and add access to the vpr data structures.
     */
    void SuccessiveOverRelaxation::BlockNode::CalculateTimingPosition(double* x_pos, double* y_pos) {
        float timing_tradeoff = -1;

        double center, maxcrit, crit;
        double wx, wy, tnewx, tnewy;
        double tmp;
        wx = wy = tnewx = tnewy = maxcrit = 0;

        if (timing_tradeoff < 0.f) {
            return;
        }

        for (size_t i = 0; i < block->num_nets(); ++i) {
            Net *net = block->nets(i);
            int pin_index = 1; // FIXME: TODO: net_pin_index[i];
            if (pin_index == 0) {  // This pin is the driver of n
                assert(block == net->block(0));
                for (size_t j = 1; j < net->num_blocks(); ++j) {
                    crit = 0; // FIXME: TODO: n->timing_place_crit(j);
                    maxcrit = std::max(maxcrit, crit);

                    center = (block->x + net->block(j)->x) * 0.5;
                    wx += tmp = crit;
                    tnewx += tmp * center;
                    center = (block->y + net->block(j)->y) * 0.5;
                    wy += tmp = crit;
                    tnewy += tmp * center;
                }
            } else if (pin_index < 0) {
                std::cerr << "Error in CalXYandNet()!" << std::endl;
                throw 1;
            }
        }

        if (wx > 0) {
            tnewx /= wx;
            tnewy /= wy;

            maxcrit *= timing_tradeoff;
            *x_pos = (*x_pos + tnewx * maxcrit) / (1 + maxcrit);
            *y_pos = (*y_pos + tnewy * maxcrit) / (1 + maxcrit);
        }
    }

    /**
     * Chapter 6.2 this applies the Jacobi iteration to each block
     * it uses the delta cost update `BlockMoved` to ensure that the
     * next block has the most up-to-date values.
     * Intuitively, this method places the block according to a weighted (by cost)
     * average of the nets which it is connected to.
     *
     * if you consider the cost to be the variance of block positions
     * then the formula is the same as weighted mean dealing with variance.
     * but the normality and independence assumptions are not necessarily
     * reasonable in this case.
     */
    void SuccessiveOverRelaxation::BlockNode::CalculatePosition(Placement* p) {
        double wx, wy, newx, newy;
        double tmp;

        wx = wy = newx = newy = 0;

        for (size_t i = 0; i < block->num_nets(); ++i) {
            Net *net = block->nets(i);
            assert(net);
            if (net->is_global) {
                continue;
            }

            wx += tmp = Net::alpha_factor / net->costx;
            assert(!isnan(wx) && !isinf(wx));
            newx += net->x * tmp;
            assert(!isnan(newx));
            wy += tmp = Net::alpha_factor / net->costy;
            assert(!isnan(wy) && !isinf(wy));
            newy += net->y * tmp;
            assert(!isnan(newy));
        }

        assert(wx > 0.0f && wy > 0.0f);
        newx /= wx;
        newy /= wy;
        assert(!isnan(newx) && !isnan(newy));

        // allow timing to influence the new position
        //CalculateTimingPosition(&newx, &newy);

        // update nets and commit to the new position
        for (size_t i = 0; i < block->num_nets(); ++i) {
            Net* net = block->nets(i);
            if (net->is_global) {
                continue;
            }
            net->BlockMoved(block->x, block->y, newx, newy);
        }

        block->x = newx;
        assert(!isnan(block->x));
        block->y = newy;
        assert(!isnan(block->y));
    }

    void SuccessiveOverRelaxation::AddToSequence(Block* block) {
        sequence.push_back(BlockLookup(block));
    }

    void SuccessiveOverRelaxation::BlockNode::setPosition(double x, double y) {
        block->x = x;
        block->y = y;
    }

    void SuccessiveOverRelaxation::NetNode::setPosition(double x, double y) {
        net->x = x;
        net->y = y;
    }

    SuccessiveOverRelaxation::BlockNode::BlockNode() {
        heap_node = new HeapNode(this);
        LPindex = -1;
        omega = 0;
        weight = 0;
    }

    SuccessiveOverRelaxation::NetNode::NetNode() {
        heap_node = new HeapNode(this);
        LPindex = -1;
        omega = 0;
        weight = 0;
    }

    SuccessiveOverRelaxation::BlockNode::~BlockNode() {
        delete heap_node;
    }

    SuccessiveOverRelaxation::NetNode::~NetNode() {
        delete heap_node;
    }

    double SuccessiveOverRelaxation::cost() {
        double cost = 0;
        for (SuccessiveOverRelaxation::netnode_iterator netnode = nets_.begin();
                netnode != nets_.end(); ++netnode) {
            (*netnode)->CalculatePosition(1.0, false, placement, blocks_, nets_);
            cost += (*netnode)->net->costx + (*netnode)->net->costy;
        }
        return Net::alpha_factor * cost;
    }
}

namespace yaml {
    std::string yaml(starplacer::SuccessiveOverRelaxation::SORNode* node) {
        if (node == NULL) {
            return "null";
        }
        std::stringstream ss;
        ss << "{";
        ss << "weight: " << node->weight;
        ss << ", ";
        ss << "omega: " << node->omega;
        ss << ", ";
        ss << "LPindex: " << node->LPindex;
        ss << ", ";
        ss << "heapWeight: " << node->heap_node->key();
        ss << "}";
        return ss.str();
    }
}
