#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <getopt.h>

#include "Netlist.h"
#include "Architecture.h"
#include "Placement.h"
#include "CGPlacer.h"
#include "SORPlacer.h"

using namespace starplacer;

typedef enum {SOR, CG} PlacementAlgorithm;

struct program_options {
    char* netlist_file;
    char* arch_file;
    char* placement_file;
    PlacementAlgorithm algo;
};

static program_options parse_options(int argc, char* const* argv);
static void show_header(const program_options& opts, const Architecture& arch, const Netlist& netlist);

int main(int argc, char** argv) {
    program_options opt = parse_options(argc, argv);

    const clock_t start = clock();
    Architecture arch(opt.arch_file);
    Netlist netlist(opt.netlist_file, arch);
    arch.init_arch(1.0f, false, netlist);

    Placement placement(netlist, arch);

    show_header(opt, arch, netlist);

    switch (opt.algo) {
        case SOR:
            SORPlacer::Place(&placement);
            break;
        case CG:
            CGPlacer::Place(&placement);
            break;
        default:
            printf("%s: error: Unknown placement algorithm\n exiting...\n", argv[0]);
            return 1;
            break;
    }

    double runtime = (clock() - start) / CLOCKS_PER_SEC;
    printf("Runtime(seconds): %.2f\n", runtime);
    placement.WriteToFile(opt.placement_file);
    return 0;
}


/**
 * Parses the command line options and set default values
 * in the program_options struct, exits if it finds any
 * invalid arguments flags.
 */
static program_options parse_options(int argc, char* const* argv) {
    program_options opts;

    // set default options
    opts.algo = SOR;

    // define optional arguments
    static struct option long_options[] = {
        {"place_algorithm", required_argument, NULL, 'a'},
        {NULL, 0, NULL, 0}};

    // parse optional arguments
    while (1) {
        int option_index = 0;
        int c = getopt_long(argc, argv, "", long_options, &option_index);

        if (c == -1) {
            break;
        }

        switch (c) {
           case 'a':
                if (strcmp("SOR", optarg) == 0) {
                    opts.algo = SOR;
                } else if (strcmp("CG", optarg) == 0) {
                    opts.algo = CG;
                } else {
                    printf("%s: error: unrecognized algorithm %s, available options "
                            "[CG, SOR, timing_driven]\n", argv[0], optarg);
                    exit(1);
                }
                break;
            default:
                abort();
        }
    }

    // parse non-optional arguments
    if (argc - optind != 3) {
        if (argc != optind) {
            printf("%s: error: incorrect number of arguments: %d\n", argv[0], argc - optind);
        }
        printf("usage: starplacer netlist.net architecture.arch placement.out "
                "[--place_algorithm SOR CG]\n");
        exit(1);
    }

    opts.netlist_file = argv[optind++];
    opts.arch_file = argv[optind++];
    opts.placement_file = argv[optind++];

    return opts;
}

static const char* algo_pretty_name(PlacementAlgorithm p) {
    switch(p) {
        case SOR:
            return "Successive Over Relaxation";
        case CG:
            return "Conjugate Gradient";
        default:
            return "Unknown Algorithm";
    }
}

static void show_header(const program_options& opts, const Architecture& arch, const Netlist& netlist) {
    printf("---#  Star+ Placer\n"
            "compiled on %s\n"
            "licensed for non-commercial use only.\n"
            "Portions of this code are from Versatile Place and Route (VPR)\n\n", __DATE__);

    printf("---#  Program Options\n");
    printf("Algorithm: %s\n", algo_pretty_name(opts.algo));

    printf("Netlist: %s\n", opts.netlist_file);
    printf("Architecture: %s\n", opts.arch_file);
    printf("Logic Blocks: %d\n", netlist.num_blocks);
    printf("Mapped onto: %d by %d logic blocks\n", arch.nx, arch.ny);
    printf("\n");
}
